import dataclasses as dc
import typing as t

import numpy
from tqdm import tqdm

from aladindb.backends.base.query import Select
from aladindb.backends.query_dataset import QueryDataset
from aladindb.base.artifact import Artifact
from aladindb.base.datalayer import Datalayer
from aladindb.components.metric import Metric
from aladindb.components.model import Model, _TrainingConfiguration


def _get_data_from_query(
    select: Select,
    X: str,
    db: Datalayer,
    y: t.Optional[str] = None,
    y_preprocess: t.Optional[t.Callable] = None,
    preprocess: t.Optional[t.Callable] = None,
):
    def transform(r):
        out = {}
        if X == '_base':
            out.update(**preprocess(r))
        else:
            out[X] = preprocess(r[X])
        if y is not None:
            out[y] = y_preprocess(r[y]) if y_preprocess else r[y]
        return out

    data = QueryDataset(
        select=select,
        keys=[X] if y is None else [X, y],
        transform=transform,
        db=db,
    )
    documents = []
    for i in tqdm(range(len(data))):
        r = data[i]
        documents.append(r)
    X_arr = [r[X] for r in documents]
    if isinstance(X[0], numpy.ndarray):
        X_arr = numpy.stack(X_arr)
    if y is not None:
        y_arr = [r[y] for r in documents]
        if y_preprocess is not None:
            y_arr = [y_preprocess(item) for item in y_arr]
        if isinstance(y[0], numpy.ndarray):
            y_arr = numpy.stack(y_arr)
    return X_arr, y_arr


@dc.dataclass
class SklearnTrainingConfiguration(_TrainingConfiguration):
    fit_params: t.Dict = dc.field(default_factory=dict)
    predict_params: t.Dict = dc.field(default_factory=dict)
    y_preprocess: t.Optional[Artifact] = None


@dc.dataclass
class Estimator(Model):
    def __post_init__(self):
        if self.predict_method is not None:
            assert self.predict_method == 'predict'
        self.predict_method = 'predict'
        super().__post_init__()

    @property
    def estimator(self):
        return self.object.artifact

    def __getattr__(self, item):
        if item in ['transform', 'predict_proba', 'score']:
            return getattr(self.estimator, item)
        else:
            return super().__getattribute__(item)

    def _forward(self, X, **kwargs):
        return self.estimator.predict(X, **kwargs)

    def _fit(  # type: ignore[override]
        self,
        X,
        y=None,
        configuration: t.Optional[SklearnTrainingConfiguration] = None,
        data_prefetch: bool = False,
        db: t.Optional[Datalayer] = None,
        metrics: t.Optional[t.Sequence[Metric]] = None,
        select: t.Optional[Select] = None,
        validation_sets: t.Optional[t.Sequence[str]] = None,
    ):
        if configuration is not None:
            self.training_configuration = configuration
        if select is not None:
            self.training_select = select
        if isinstance(X, str):
            self.train_X = X
            self.train_y = y
        if isinstance(X, str):
            if isinstance(self.training_configuration, str):
                raise TypeError('self.training_configuration cannot be str')
            y_preprocess = None
            if self.training_configuration is not None:
                y_preprocess = self.training_configuration.get('y_preprocess', None)
                if isinstance(y_preprocess, Artifact):
                    y_preprocess = y_preprocess.artifact
            if select is None or db is None:
                raise ValueError('Neither select nor db can be None')

            if isinstance(self.preprocess, Artifact):
                preprocess = self.preprocess.artifact
            else:

                def preprocess(x):
                    return x

            X, y = _get_data_from_query(
                select=select,
                db=db,
                X=X,
                y=y,
                preprocess=preprocess,
                y_preprocess=y_preprocess,
            )
        if self.training_configuration is not None:
            assert not isinstance(self.training_configuration, str)
            to_return = self.estimator.fit(
                X=X,
                y=y,
                **self.training_configuration.get('fit_params', {}),
            )
        else:
            to_return = self.estimator.fit(X, y)

        if validation_sets is not None:
            assert db is not None
            assert metrics is not None
            results = {}
            for validation_set in validation_sets:
                results.update(
                    # TODO: this call could never work
                    self._validate(
                        db=db,
                        validation_set=validation_set,
                        metrics=metrics,
                    )
                )
            self.append_metrics(results)
        if db is not None:
            db.replace(self, upsert=True)
        return to_return



from aladindb.base.build import build_datalayer
from aladindb.components.stack import Stack

from . import command


@command(help='Apply the stack tarball to the database')
def apply(yaml_path: str, identifier: str):
    db = build_datalayer()
    stack = Stack(identifier=identifier)
    stack.load(yaml_path)
    db.add(stack)



"""
Change Data Capture (CDC) is a mechanism used in database systems to track
and capture changes made to a table or collection in real-time.
It allows applications to stay up-to-date with the latest changes in the database
and perform various tasks, such as data synchronization, auditing,
or data integration. The ChangeDataCapture class is designed to
handle CDC tasksfor a specified table/collection in a database.

Change streams allow applications to access real-time data changes
without the complexity and risk of tailing the oplog.
Applications can use change streams to subscribe to all data
changes on a single collection,a database, or an entire deployment,
and immediately react to them.
Because change streams use the aggregation framework, applications
can also filter for specific changes or transform the notifications at will.

ref: https://www.mongodb.com/docs/manual/changeStreams/

Use this module like this::
    db = any_arbitary_database.connect(...)
    db = aladin(db)
    db.cdc.start()
    # or
    db.cdc.listen(on=Collection('test_collection'))
"""

import dataclasses as dc
import json
import queue
import threading
import traceback
import typing as t
from abc import ABC, abstractmethod
from collections import Counter
from enum import Enum

from pymongo.change_stream import CollectionChangeStream

from aladindb import logging
from aladindb.misc.runnable.queue_chunker import QueueChunker
from aladindb.misc.runnable.runnable import Event

if t.TYPE_CHECKING:
    from aladindb.backends.base.query import TableOrCollection
    from aladindb.base.datalayer import Datalayer
    from aladindb.base.serializable import Serializable
    from aladindb.components.listener import Listener


class DBEvent(str, Enum):
    """`DBEvent` simple enum to hold mongo basic events."""

    delete = 'delete'
    insert = 'insert'
    update = 'update'


@dc.dataclass
class Packet:
    ids: t.Any
    query: t.Optional['Serializable']
    event_type: DBEvent = DBEvent.insert

    @property
    def is_delete(self) -> bool:
        return self.event_type == DBEvent.delete

    @staticmethod
    def collate(packets: t.Sequence['Packet']) -> 'Packet':
        """
        Collate a batch of packets into one
        """
        assert packets
        ids = [packet.ids[0] for packet in packets]
        query = packets[0].query

        # TODO: cluster Packet for each event.
        event_type = packets[0].event_type
        return type(packets[0])(ids=ids, query=query, event_type=event_type)


queue_chunker = QueueChunker(chunk_size=100, timeout=0.2)


class BaseDatabaseListener(ABC):
    """
    A Base class which defines basic functions to implement.
    """

    IDENTITY_SEP: str = '/'
    _scheduler: t.Optional[threading.Thread]
    Packet: Packet

    def __init__(
        self,
        db: 'Datalayer',
        on: 'TableOrCollection',
        stop_event: Event,
        identifier: 'str' = '',
        timeout: t.Optional[float] = None,
    ):
        self.db = db
        self._on_component = on
        self._change_counters = Counter(inserts=0, updates=0, deletes=0)
        self._identifier = self._build_identifier([identifier, on.identifier])
        self._stop_event = stop_event
        self._startup_event = Event()
        self._scheduler = None
        self.timeout = timeout
        self.db_type = None

        from aladindb.backends.base import backends
        from aladindb.backends.mongodb.cdc.base import MongoDBPacket

        if isinstance(self.db.databackend, backends.MongoDataBackend):
            self.db_type = 'mongodb'
            self.packet = lambda ids, query, event_type: MongoDBPacket(
                ids, query, event_type
            )
        else:
            raise NotImplementedError(f'{self.db.databackend} not supported yet!')

    @property
    def identity(self) -> str:
        return self._identifier

    @classmethod
    def _build_identifier(cls, identifiers) -> str:
        """_build_identifier.

        :param identifiers: list of identifiers.
        :rtype: str
        """
        return cls.IDENTITY_SEP.join(identifiers)

    def info(self) -> t.Dict:
        """
        Get info on the current state of listener.
        """
        info = {}
        info.update(
            {
                'inserts': self._change_counters['inserts'],
                'updates': self._change_counters['updates'],
                'deletes': self._change_counters['deletes'],
            }
        )
        logging.info(json.dumps(info, indent=2))
        return info

    @abstractmethod
    def listen(self):
        raise NotImplementedError

    @abstractmethod
    def stop(self):
        raise NotImplementedError

    @abstractmethod
    def setup_cdc(self) -> CollectionChangeStream:
        raise NotImplementedError

    @abstractmethod
    def on_create(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def on_update(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def on_delete(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def next_cdc(self, stream: CollectionChangeStream) -> None:
        raise NotImplementedError

    def create_event(
        self,
        ids: t.Sequence,
        db: 'Datalayer',
        table_or_collection: 'TableOrCollection',
        event: DBEvent,
    ):
        """
        A helper to create packet based on the event type and put it on the cdc queue

        :param change: The changed document.
        :param db: a aladindb instance.
        :param table_or_collection: The collection on which change was observed.
        :param event: CDC event type
        """
        cdc_query = None
        # TODO why was this logic here? Why is the query always the same?
        # if event != DBEvent.delete:
        cdc_query = table_or_collection.find()

        db.cdc.CDC_QUEUE.put_nowait(self.packet(ids, cdc_query, event))

    def event_handler(self, ids: t.Sequence, event: DBEvent) -> None:
        """event_handler.
        A helper fxn to handle incoming changes from change stream on a collection.

        :param ids: Changed document ids
        :param event: CDC event
        """
        if event == DBEvent.insert:
            self._change_counters['inserts'] += 1
            self.on_create(ids, db=self.db, collection=self._on_component)

        elif event == DBEvent.update:
            self._change_counters['updates'] += 1
            self.on_update(ids, db=self.db, collection=self._on_component)

        elif event == DBEvent.delete:
            self._change_counters['deletes'] += 1
            self.on_delete(ids, db=self.db, collection=self._on_component)


class DatabaseListenerThreadScheduler(threading.Thread):
    def __init__(
        self,
        listener: BaseDatabaseListener,
        stop_event: Event,
        start_event: Event,
    ) -> None:
        threading.Thread.__init__(self, daemon=True)
        self.stop_event = stop_event
        self.start_event = start_event
        self.listener = listener

    def run(self) -> None:
        try:
            self.listener.db.rebuild()
            cdc_stream = self.listener.setup_cdc()
            self.start_event.set()
            while not self.stop_event.is_set():
                self.listener.next_cdc(cdc_stream)
        except Exception as e:
            logging.error('In DatabaseListenerThreadScheduler', str(e))
            traceback.print_exc()


class CDCHandler(threading.Thread):
    """
    This class is responsible for handling the change by executing the taskflow.
    This class also extends the task graph by adding funcation job node which
    does post model executiong jobs, i.e `copy_vectors`.
    """

    def __init__(self, db: 'Datalayer', stop_event: Event, queue):
        """__init__.

        :param db: a aladindb instance.
        :param stop_event: A threading event flag to notify for stoppage.
        """
        db.rebuild()

        self.db = db
        self._stop_event = stop_event
        self._is_running = False
        self.cdc_queue = queue
        threading.Thread.__init__(self, daemon=True)

    @property
    def is_running(self):
        return self._is_running

    def run(self):
        self._is_running = True
        try:
            for c in queue_chunker(self.cdc_queue, self._stop_event):
                packet = Packet.collate(c)
                if packet.is_delete:
                    self.db.refresh_after_delete(packet.query, packet.ids)
                else:
                    self.db.refresh_after_update_or_insert(
                        packet.query,
                        packet.ids,
                    )

        except Exception as exc:
            logging.error("Error while handling cdc batches :: reason", exc)
            traceback.print_exc()
        finally:
            self._is_running = False


DBListenerType = t.TypeVar('DBListenerType')


class DatabaseListenerFactory(t.Generic[DBListenerType]):
    """A Factory class to create instance of DatabaseListener corresponding to the
    `db_type`.
    """

    SUPPORTED_LISTENERS: t.List[str] = ['mongodb']

    def __init__(self, db_type: str = 'mongodb'):
        if db_type not in self.SUPPORTED_LISTENERS:
            raise ValueError(f'{db_type} is not supported yet for CDC.')
        self.db_type = db_type

    def create(self, *args, **kwargs) -> DBListenerType:
        stop_event = Event()
        kwargs['stop_event'] = stop_event
        if self.db_type == 'mongodb':
            from aladindb.backends.mongodb.cdc.listener import MongoDatabaseListener

            listener = MongoDatabaseListener(*args, **kwargs)
            return t.cast(DBListenerType, listener)
        else:
            raise NotImplementedError


class DatabaseChangeDataCapture:
    """
    DatabaseChangeDataCapture is a Python class that provides a flexible and
    extensible framework for capturing and managing data changes
    in a database.

    This class is repsonsible for cdc service on the provided `db` instance
    This class is designed to simplify the process of tracking changes
    to database records,allowing you to monitor and respond to
    data modifications efficiently.
    """

    def __init__(self, db: 'Datalayer'):
        self.db = db
        self._cdc_stop_event = Event()
        self.CDC_QUEUE: queue.Queue = queue.Queue()
        self.cdc_change_handler: t.Optional[CDCHandler] = None
        self._CDC_LISTENERS: t.Dict[str, BaseDatabaseListener] = {}
        self._running: bool = False
        self._cdc_existing_collections: t.MutableSequence['TableOrCollection'] = []

        listeners = self.db.show('listeners')
        if listeners:
            from aladindb.components.listener import Listener

            for listener in listeners:
                listener = db.load(identifier=listener, type_id='listener')
                assert isinstance(listener, Listener)
                self.add(listener)

    @property
    def running(self) -> bool:
        return self._running

    def start(self):
        """
        This method starts the cdc process on the database.
        """
        self._running = True

        # listen to existing collection without cdc enabled
        for collection in self._cdc_existing_collections:
            self.listen(collection)

    def listen(
        self,
        on: 'TableOrCollection',
        identifier: str = '',
        *args,
        **kwargs,
    ):
        """
        Starts cdc service on the provided collection
        Not to be confused with ``aladindb.container.listener.Listener``.

        :param db: A aladindb instance.
        :param on: Which collection/table listener service this be invoked on?
        :param identifier: A identity given to the listener service.
        """
        from aladindb.backends.base import backends

        if isinstance(self.db.databackend, backends.MongoDataBackend):
            db_type = 'mongodb'
        else:
            raise NotImplementedError(f'{self.db.databackend} not supported yet!')

        if not self.cdc_change_handler:
            cdc_change_handler = CDCHandler(
                db=self.db, stop_event=self._cdc_stop_event, queue=self.CDC_QUEUE
            )
            cdc_change_handler.start()
            self.cdc_change_handler = cdc_change_handler

        db_factory: DatabaseListenerFactory = DatabaseListenerFactory(db_type=db_type)
        listener = db_factory.create(
            db=self.db, on=on, identifier=identifier, *args, **kwargs
        )
        self._CDC_LISTENERS[on.identifier] = listener  # type: ignore[index]

        listener.listen()
        return listener

    def stop(self, name: str = ''):
        """
        Stop all registered listeners
        :param name: Listener name
        """
        try:
            if name:
                try:
                    self._CDC_LISTENERS[name].stop()
                except KeyError:
                    raise KeyError(f'{name} is already down or not added yet')
                else:
                    del self._CDC_LISTENERS[name]

            for _, listener in self._CDC_LISTENERS.items():
                listener.stop()
        finally:
            self._running = False
            self._CDC_LISTENERS = {}
            self.stop_handler()

    def stop_handler(self):
        """
        Stop the cdc handler thread
        """
        self._cdc_stop_event.set()
        if self.cdc_change_handler:
            self.cdc_change_handler.join()
        self.cdc_change_handler = None

    def add(self, listener: 'Listener'):
        """
        This method registered the given collection for cdc
        """
        collection = listener.select.table_or_collection
        if self.running and collection.identifier not in self._CDC_LISTENERS:
            self.listen(collection)
        else:
            # Append to existing collection list
            self._cdc_existing_collections.append(collection)



import asyncio
import base64
import dataclasses as dc
import itertools
import os
import typing as t

import aiohttp
import requests
import tqdm
from openai import (
    APITimeoutError,
    AsyncOpenAI,
    InternalServerError,
    OpenAI as SyncOpenAI,
    RateLimitError,
)

from aladindb.backends.ibis.data_backend import IbisDataBackend
from aladindb.backends.ibis.field_types import dtype
from aladindb.base.datalayer import Datalayer
from aladindb.components.model import APIModel
from aladindb.components.vector_index import sqlvector, vector
from aladindb.misc.compat import cache
from aladindb.misc.retry import Retry

retry = Retry(exception_types=(RateLimitError, InternalServerError, APITimeoutError))


@cache
def _available_models():
    return tuple([r.id for r in SyncOpenAI().models.list().data])


@dc.dataclass(kw_only=True)
class _OpenAI(APIModel):
    __doc__ = APIModel.__doc__  # type: ignore[assignment]

    def __post_init__(self):
        super().__post_init__()

        # dall-e is not currently included in list returned by OpenAI model endpoint
        if self.model not in (mo := _available_models()) and self.model not in (
            'dall-e'
        ):
            msg = f'model {self.model} not in OpenAI available models, {mo}'
            raise ValueError(msg)

        self.syncClient = SyncOpenAI()
        self.asyncClient = AsyncOpenAI()

        if 'OPENAI_API_KEY' not in os.environ:
            raise ValueError('OPENAI_API_KEY not set')


@dc.dataclass(kw_only=True)
class OpenAIEmbedding(_OpenAI):
    """
    OpenAI embedding predictor
    {_openai_parameters}
    :param shape: The shape as ``tuple`` of the embedding.
    """

    __doc__ = __doc__.format(_openai_parameters=_OpenAI.__doc__)

    shape: t.Optional[t.Sequence[int]] = None
    shapes: t.ClassVar[t.Dict] = {'text-embedding-ada-002': (1536,)}

    def __post_init__(self):
        super().__post_init__()
        if self.shape is None:
            self.shape = self.shapes[self.identifier]

    def pre_create(self, db):
        super().pre_create(db)
        if isinstance(db.databackend, IbisDataBackend):
            if self.encoder is None:
                self.encoder = sqlvector(self.shape)
        elif self.encoder is None:
            self.encoder = vector(self.shape)

    @retry
    def _predict_one(self, X: str, **kwargs):
        e = self.syncClient.embeddings.create(input=X, model=self.identifier, **kwargs)
        return e.data[0].embedding

    @retry
    async def _apredict_one(self, X: str, **kwargs):
        e = await self.asyncClient.embeddings.create(
            input=X, model=self.identifier, **kwargs
        )
        return e.data[0].embedding

    @retry
    def _predict_a_batch(self, texts: t.List[str], **kwargs):
        out = self.syncClient.embeddings.create(
            input=texts, model=self.identifier, **kwargs
        )
        return [r.embedding for r in out.data]

    @retry
    async def _apredict_a_batch(self, texts: t.List[str], **kwargs):
        out = await self.asyncClient.embeddings.create(
            input=texts, model=self.identifier, **kwargs
        )
        return [r.embedding for r in out.data]

    def _predict(self, X, one: bool = False, **kwargs):
        if isinstance(X, str):
            return self._predict_one(X)
        out = []
        batch_size = kwargs.pop('batch_size', 100)
        for i in tqdm.tqdm(range(0, len(X), batch_size)):
            out.extend(self._predict_a_batch(X[i : i + batch_size], **kwargs))
        return out

    async def _apredict(self, X, one: bool = False, **kwargs):
        if isinstance(X, str):
            return await self._apredict_one(X)
        out = []
        batch_size = kwargs.pop('batch_size', 100)
        # Note: we submit the async requests in serial to avoid rate-limiting
        for i in range(0, len(X), batch_size):
            out.extend(await self._apredict_a_batch(X[i : i + batch_size], **kwargs))
        return out


@dc.dataclass(kw_only=True)
class OpenAIChatCompletion(_OpenAI):
    """OpenAI chat completion predictor.
    {_openai_parameters}
    :param prompt: The prompt to use to seed the response.
    """

    __doc__ = __doc__.format(_openai_parameters=_OpenAI.__doc__)

    prompt: str = ''

    def __post_init__(self):
        super().__post_init__()
        self.takes_context = True

    def _format_prompt(self, context, X):
        prompt = self.prompt.format(context='\n'.join(context))
        return prompt + X

    def pre_create(self, db: Datalayer) -> None:
        super().pre_create(db)
        if isinstance(db.databackend, IbisDataBackend) and self.encoder is None:
            self.encoder = dtype('str')

    @retry
    def _predict_one(self, X, context: t.Optional[t.List[str]] = None, **kwargs):
        if context is not None:
            X = self._format_prompt(context, X)
        return (
            self.syncClient.chat.completions.create(
                messages=[{'role': 'user', 'content': X}],
                model=self.identifier,
                **kwargs,
            )
            .choices[0]
            .message.content
        )

    @retry
    async def _apredict_one(self, X, context: t.Optional[t.List[str]] = None, **kwargs):
        if context is not None:
            X = self._format_prompt(context, X)
        return (
            (
                await self.asyncClient.chat.completions.create(
                    messages=[{'role': 'user', 'content': X}],
                    model=self.identifier,
                    **kwargs,
                )
            )
            .choices[0]
            .message.content
        )

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(X, context=context, **kwargs)
        return [self._predict_one(msg) for msg in X]

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(X, context=context, **kwargs)
        return [await self._apredict_one(msg) for msg in X]


@dc.dataclass(kw_only=True)
class OpenAIImageCreation(_OpenAI):
    """OpenAI image creation predictor.
    {_openai_parameters}
    :param takes_context: Whether the model takes context into account.
    :param prompt: The prompt to use to seed the response.
    """

    __doc__ = __doc__.format(_openai_parameters=_OpenAI.__doc__)

    takes_context: bool = True
    prompt: str = ''

    def pre_create(self, db: Datalayer) -> None:
        super().pre_create(db)
        if isinstance(db.databackend, IbisDataBackend) and self.encoder is None:
            self.encoder = dtype('bytes')

    def _format_prompt(self, context, X):
        prompt = self.prompt.format(context='\n'.join(context))
        return prompt + X

    @retry
    def _predict_one(
        self,
        X,
        n: int,
        response_format: str,
        context: t.Optional[t.List[str]] = None,
        **kwargs,
    ):
        if context is not None:
            X = self._format_prompt(context, X)
        if response_format == 'b64_json':
            b64_json = (
                self.syncClient.images.generate(
                    prompt=X, n=n, response_format='b64_json'
                )
                .data[0]
                .b64_json
            )
            return base64.b64decode(b64_json)
        else:
            url = self.syncClient.images.generate(prompt=X, n=n, **kwargs).data[0].url
            return requests.get(url).content

    @retry
    async def _apredict_one(
        self,
        X,
        n: int,
        response_format: str,
        context: t.Optional[t.List[str]] = None,
        **kwargs,
    ):
        if context is not None:
            X = self._format_prompt(context, X)
        if response_format == 'b64_json':
            b64_json = (
                (
                    await self.asyncClient.images.generate(
                        prompt=X, n=n, response_format='b64_json'
                    )
                )
                .data[0]
                .b64_json
            )
            return base64.b64decode(b64_json)
        else:
            url = (
                (await self.asyncClient.images.generate(prompt=X, n=n, **kwargs))
                .data[0]
                .url
            )
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as resp:
                    return await resp.read()

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        response_format = kwargs.pop('response_format', 'b64_json')
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(
                X, n=1, response_format=response_format, context=context, **kwargs
            )
        return [
            self._predict_one(msg, n=1, response_format=response_format) for msg in X
        ]

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        response_format = kwargs.pop('response_format', 'b64_json')
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(
                X, context=context, n=1, response_format=response_format, **kwargs
            )
        return [
            await self._apredict_one(msg, n=1, response_format=response_format)
            for msg in X
        ]


@dc.dataclass(kw_only=True)
class OpenAIImageEdit(_OpenAI):
    """OpenAI image edit predictor.
    {_openai_parameters}
    :param takes_context: Whether the model takes context into account.
    :param prompt: The prompt to use to seed the response.
    """

    __doc__ = __doc__.format(_openai_parameters=_OpenAI.__doc__)

    takes_context: bool = True
    prompt: str = ''

    def _format_prompt(self, context):
        prompt = self.prompt.format(context='\n'.join(context))
        return prompt

    def pre_create(self, db: Datalayer) -> None:
        super().pre_create(db)
        if isinstance(db.databackend, IbisDataBackend) and self.encoder is None:
            self.encoder = dtype('bytes')

    @retry
    def _predict_one(
        self,
        image: t.BinaryIO,
        n: int,
        response_format: str,
        context: t.Optional[t.List[str]] = None,
        mask_png_path: t.Optional[str] = None,
        **kwargs,
    ):
        if context is not None:
            self.prompt = self._format_prompt(context)

        if mask_png_path is not None:
            with open(mask_png_path, 'rb') as f:
                mask = f.read()
            kwargs['mask'] = mask

        if response_format == 'b64_json':
            b64_json = (
                self.syncClient.images.edit(
                    image=image,
                    prompt=self.prompt,
                    n=n,
                    response_format='b64_json',
                    **kwargs,
                )
                .data[0]
                .b64_json
            )
            return base64.b64decode(b64_json)
        else:
            url = (
                self.syncClient.images.edit(
                    image=image, prompt=self.prompt, n=n, **kwargs
                )
                .data[0]
                .url
            )
            return requests.get(url).content

    @retry
    async def _apredict_one(
        self,
        image: t.BinaryIO,
        n: int,
        response_format: str,
        context: t.Optional[t.List[str]] = None,
        mask_png_path: t.Optional[str] = None,
        **kwargs,
    ):
        if context is not None:
            self.prompt = self._format_prompt(context)

        if mask_png_path is not None:
            with open(mask_png_path, 'rb') as f:
                mask = f.read()
            kwargs['mask'] = mask

        if response_format == 'b64_json':
            b64_json = (
                (
                    await self.asyncClient.images.edit(
                        image=image,
                        prompt=self.prompt,
                        n=n,
                        response_format='b64_json',
                        **kwargs,
                    )
                )
                .data[0]
                .b64_json
            )
            return base64.b64decode(b64_json)
        else:
            url = (
                (
                    await self.asyncClient.images.edit(
                        image=image, prompt=self.prompt, n=n, **kwargs
                    )
                )
                .data[0]
                .url
            )
            async with aiohttp.ClientSession() as session:
                async with session.get(url) as resp:
                    return await resp.read()

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        response_format = kwargs.pop('response_format', 'b64_json')
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(
                image=X, n=1, response_format=response_format, context=context, **kwargs
            )
        return [
            self._predict_one(
                image=image, n=1, response_format=response_format, **kwargs
            )
            for image in X
        ]

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        response_format = kwargs.pop('response_format', 'b64_json')
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(
                image=X, context=context, n=1, response_format=response_format, **kwargs
            )
        return [
            await self._apredict_one(
                image=image, n=1, response_format=response_format, **kwargs
            )
            for image in X
        ]


@dc.dataclass(kw_only=True)
class OpenAIAudioTranscription(_OpenAI):
    """OpenAI audio transcription predictor.
    {_openai_parameters}
    :param takes_context: Whether the model takes context into account.
    :param prompt: The prompt to guide the model's style. Should contain ``{context}``.
    """

    __doc__ = __doc__.format(_openai_parameters=_OpenAI.__doc__, context='{context}')

    takes_context: bool = True
    prompt: str = ''

    def pre_create(self, db: Datalayer) -> None:
        super().pre_create(db)
        if isinstance(db.databackend, IbisDataBackend) and self.encoder is None:
            self.encoder = dtype('str')

    @retry
    def _predict_one(
        self, file: t.BinaryIO, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        "Converts a file-like Audio recording to text."
        if context is not None:
            self.prompt = self.prompt.format(context='\n'.join(context))
        return self.syncClient.audio.transcriptions.create(
            file=file,
            model=self.identifier,
            prompt=self.prompt,
            **kwargs,
        ).text

    @retry
    async def _apredict_one(
        self, file: t.BinaryIO, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        "Converts a file-like Audio recording to text."
        if context is not None:
            self.prompt = self.prompt.format(context='\n'.join(context))
        return (
            await self.asyncClient.audio.transcriptions.create(
                file=file,
                model=self.identifier,
                prompt=self.prompt,
                **kwargs,
            )
        ).text

    @retry
    def _predict_a_batch(self, files: t.List[t.BinaryIO], **kwargs):
        "Converts multiple file-like Audio recordings to text."
        resps = [
            self.syncClient.audio.transcriptions.create(
                file=file, model=self.identifier, **kwargs
            )
            for file in files
        ]
        return [resp.text for resp in resps]

    @retry
    async def _apredict_a_batch(self, files: t.List[t.BinaryIO], **kwargs):
        "Converts multiple file-like Audio recordings to text."
        resps = await asyncio.gather(
            *[
                self.asyncClient.audio.transcriptions.create(
                    file=file, model=self.identifier, **kwargs
                )
                for file in files
            ]
        )
        return [resp.text for resp in resps]

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(X, context=context, **kwargs)
        out = []
        batch_size = kwargs.pop('batch_size', 10)
        for i in tqdm.tqdm(range(0, len(X), batch_size)):
            out.extend(self._predict_a_batch(X[i : i + batch_size], **kwargs))
        return out

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(X, context=context, **kwargs)
        batch_size = kwargs.pop('batch_size', 10)
        list_of_lists = await asyncio.gather(
            *[
                self._apredict_a_batch(X[i : i + batch_size], **kwargs)
                for i in range(0, len(X), batch_size)
            ]
        )
        return list(itertools.chain(*list_of_lists))


@dc.dataclass(kw_only=True)
class OpenAIAudioTranslation(_OpenAI):
    """OpenAI audio translation predictor.
    {_openai_parameters}
    :param takes_context: Whether the model takes context into account.
    :param prompt: The prompt to guide the model's style. Should contain ``{context}``.
    """

    __doc__ = __doc__.format(_openai_parameters=_OpenAI.__doc__, context='{context}')

    takes_context: bool = True
    prompt: str = ''

    def pre_create(self, db: Datalayer) -> None:
        super().pre_create(db)
        if isinstance(db.databackend, IbisDataBackend) and self.encoder is None:
            self.encoder = dtype('str')

    @retry
    def _predict_one(
        self, file: t.BinaryIO, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        "Translates a file-like Audio recording to English."
        if context is not None:
            self.prompt = self.prompt.format(context='\n'.join(context))
        return (
            self.syncClient.audio.translations.create(
                file=file,
                model=self.identifier,
                prompt=self.prompt,
                **kwargs,
            )
        ).text

    @retry
    async def _apredict_one(
        self, file: t.BinaryIO, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        "Translates a file-like Audio recording to English."
        if context is not None:
            self.prompt = self.prompt.format(context='\n'.join(context))
        return (
            await self.asyncClient.audio.translations.create(
                file=file,
                model=self.identifier,
                prompt=self.prompt,
                **kwargs,
            )
        ).text

    @retry
    def _predict_a_batch(self, files: t.List[t.BinaryIO], **kwargs):
        "Translates multiple file-like Audio recordings to English."
        resps = [
            self.syncClient.audio.translations.create(
                file=file, model=self.identifier, **kwargs
            )
            for file in files
        ]
        return [resp.text for resp in resps]

    @retry
    async def _apredict_a_batch(self, files: t.List[t.BinaryIO], **kwargs):
        "Translates multiple file-like Audio recordings to English."
        resps = await asyncio.gather(
            *[
                self.asyncClient.audio.translations.create(
                    file=file, model=self.identifier, **kwargs
                )
                for file in files
            ]
        )
        return [resp.text for resp in resps]

    def _predict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return self._predict_one(X, context=context, **kwargs)
        out = []
        batch_size = kwargs.pop('batch_size', 10)
        for i in tqdm.tqdm(range(0, len(X), batch_size)):
            out.extend(self._predict_a_batch(X[i : i + batch_size], **kwargs))
        return out

    async def _apredict(
        self, X, one: bool = True, context: t.Optional[t.List[str]] = None, **kwargs
    ):
        if context:
            assert one, 'context only works with ``one=True``'
        if one:
            return await self._apredict_one(X, context=context, **kwargs)
        batch_size = kwargs.pop('batch_size', 10)
        list_of_lists = await asyncio.gather(
            *[
                self._apredict_a_batch(X[i : i + batch_size], **kwargs)
                for i in range(0, len(X), batch_size)
            ]
        )
        return list(itertools.chain(*list_of_lists))



from __future__ import annotations

import enum
import typing as t
from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import numpy
import numpy.typing

if t.TYPE_CHECKING:
    from aladindb.components.vector_index import VectorIndex


class BaseVectorSearcher(ABC):
    @classmethod
    def from_component(cls, vi: 'VectorIndex'):
        return cls(
            identifier=vi.identifier, dimensions=vi.dimensions, measure=vi.measure
        )

    @abstractmethod
    def __init__(
        self,
        identifier: str,
        dimensions: int,
        h: t.Optional[numpy.ndarray] = None,
        index: t.Optional[t.List[str]] = None,
        measure: t.Optional[str] = None,
    ):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @staticmethod
    def to_numpy(h):
        if isinstance(h, numpy.ndarray):
            return h
        if hasattr(h, 'numpy'):
            return h.numpy()
        if isinstance(h, list):
            return numpy.array(h)
        raise ValueError(str(h))

    @staticmethod
    def to_list(h):
        if hasattr(h, 'tolist'):
            return h.tolist()
        if isinstance(h, list):
            return h
        raise ValueError(str(h))

    @abstractmethod
    def add(self, items: t.Sequence[VectorItem]) -> None:
        """
        Add items to the index.

        :param items: t.Sequence of VectorItems
        """

    @abstractmethod
    def delete(self, ids: t.Sequence[str]) -> None:
        """
        Remove items from the index

        :param ids: t.Sequence of ids of vectors.
        """

    @abstractmethod
    def find_nearest_from_id(
        self,
        _id,
        n: int = 100,
        within_ids: t.Sequence[str] = (),
    ) -> t.Tuple[t.List[str], t.List[float]]:
        """
        Find the nearest vectors to the vector with the given id.

        :param _id: id of the vector
        :param n: number of nearest vectors to return
        """

    @abstractmethod
    def find_nearest_from_array(
        self,
        h: numpy.typing.ArrayLike,
        n: int = 100,
        within_ids: t.Sequence[str] = (),
    ) -> t.Tuple[t.List[str], t.List[float]]:
        """
        Find the nearest vectors to the given vector.

        :param h: vector
        :param n: number of nearest vectors to return
        """


class VectorIndexMeasureType(str, enum.Enum):
    cosine = 'cosine'
    css = 'css'
    dot = 'dot'
    l2 = 'l2'


@dataclass(frozen=True)
class VectorSearchConfig:
    '''
    Represents search config which helps initiate a vector
    searcher class.
    '''

    id: str
    dimensions: int
    measure: VectorIndexMeasureType = VectorIndexMeasureType.l2
    parameters: t.Mapping[str, t.Any] = field(default_factory=dict)


@dataclass(frozen=True)
class VectorItem:
    '''
    Class for representing a vector in vector search with
    id and vector.

    '''

    id: str
    vector: numpy.ndarray

    @classmethod
    def create(
        cls,
        *,
        id: str,
        vector: numpy.typing.ArrayLike,
    ) -> VectorItem:
        return VectorItem(id=id, vector=BaseVectorSearcher.to_numpy(vector))

    def to_dict(self) -> t.Dict:
        return {'id': self.id, 'vector': self.vector}


@dataclass(frozen=True)
class VectorSearchResult:
    '''
    Dataclass for representing vector search results with
    `id` and `score`.
    '''

    id: str
    score: float


def l2(x, y):
    '''
    L2 function for vector similarity search
    '''
    return numpy.array([-numpy.linalg.norm(x - y, axis=1)])


def dot(x, y):
    '''
    Dot function for vector similarity search
    '''
    return numpy.dot(x, y.T)


def cosine(x, y):
    '''
    Cosine similarity function for vector search
    '''
    x = x / numpy.linalg.norm(x, axis=1)[:, None]
    y = y / numpy.linalg.norm(y, axis=1)[:, None]
    return dot(x, y)


measures = {'cosine': cosine, 'dot': dot, 'l2': l2}



import dataclasses as dc
import typing as t

import aladindb as s


def for_each(
    fn: t.Callable[[t.Any], t.Any], item: t.Any, depth_first: bool = False
) -> None:
    """
    Recursively applies a function, breadth-first.

    `for_each` recurses through dicts, lists and tuples, and also through
    members of dataclasses and Pydantic model.

    :param fn: function to apply
    :param item: item to apply function to
    :param depth_first: whether to apply the function before recursing
    """
    if not depth_first:
        fn(item)

    it: t.Iterable = ()

    if isinstance(item, (list, tuple)):
        it = item
    elif isinstance(item, dict):
        it = item.values()
    elif dc.is_dataclass(item):
        it = (getattr(item, f.name) for f in dc.fields(item))
    elif isinstance(item, s.JSONable):
        it = (getattr(item, i) for i in item.schema()['properties'])

    for i in it:
        for_each(fn, i)

    if depth_first:
        fn(item)



import typing as t

"""
Utilities to deal with trees made up recursively of dicts and lists,
with leaves of any type.
"""

Accept = t.Callable[[t.Any], bool]
Rewrite = t.Callable[[t.Any], t.Any]


def tree_find(tree: t.Any, accept: Accept) -> t.Iterator[t.Any]:
    """Iterates recursively through lists and dicts, yielding all leaves `x`
       where `accept(x)` is True.

    :param tree: A tree made of dicts and lists
    :param accept: A function which returns True for values to yield
    """
    if accept(tree):
        yield tree
        return

    if isinstance(tree, dict):
        it = iter(tree.values())
    elif isinstance(tree, list):
        it = iter(tree)
    else:
        return

    yield from (a for i in it for a in tree_find(i, accept))


def tree_rewrite(tree: t.Any, accept: Accept, rewrite: Rewrite) -> t.Any:
    """Iterates recursively through lists and dicts, rewriting all leaves `x`
       where `accept(x)` is True with `rewrite(x)`

    :param tree: A tree made of dicts and lists
    :param accept: A function which returns True for values to rewrite
    :param rewrite: A function which rewrites accepted values
    """
    if accept(tree):
        return rewrite(tree)
    if isinstance(tree, list):
        return [tree_rewrite(t, accept, rewrite) for t in tree]
    if isinstance(tree, dict):
        return {k: tree_rewrite(v, accept, rewrite) for k, v in tree.items()}
    return tree



import dataclasses as dc
import traceback
import typing as t
from functools import cached_property, partial, wraps
from threading import Thread

import aladindb as s

from .runnable import Callback, Runnable


def none(x):
    pass


def _debug(method=None, before=True, after=False):
    if method is None:
        return partial(_debug, before=before, after=after)

    @wraps(method)
    def wrapped(self, *a, **ka):
        msg = f'{self}: {method.__name__}'

        try:
            before and s.logging.debug(f'{msg}: before')
            return method(self, *a, **ka)
        finally:
            after and s.logging.debug(f'{msg}: after')

    return wrapped


class ThreadBase(Runnable):
    """A base class for classes with a thread.

    It adds the following features to threading.Thread:

    * Has Events `running` and `stopped` with `on_set` callbacks
    * Handles exceptions and prints or redirects them
    * Runs once, or multiple times, depending on `self.looping`
    """

    callback: Callback
    error: t.Callable[[Exception], None]
    daemon: bool = False
    looping: bool = False
    name: str = ''

    def __str__(self):
        return f'({self.__class__.__name__}){self.name}'

    @_debug
    def pre_run(self):
        pass

    @_debug(after=True)
    def run(self):
        self.pre_run()
        self.running.set()

        while self.running:
            try:
                self.callback()
            except Exception as e:
                exc = traceback.format_exc()
                s.logging.error(f'{self}: Exception\n{exc}')

                self.error(e)
                self.stop()
            else:
                if not self.looping:
                    break
        if self.running:
            self.running.clear()
        self.stopped.set()

    @_debug
    def stop(self):
        self.running.clear()

    @_debug
    def finish(self):
        pass


class IsThread(ThreadBase, Thread):
    """This ThreadBase inherits from threading.Thread.

    To use IsThread, derive from it and override either or both of
    self.callback() and self.pre_run()
    """

    def __init__(self, *args, **kwargs):
        ThreadBase.__init__(self, *args, **kwargs)
        Thread.__init__(self, daemon=self.daemon)

    def callback(self):
        pass

    def error(self, item: Exception) -> None:
        pass

    @_debug(after=True)
    def join(self, timeout: t.Optional[float] = None):
        Thread.join(self, timeout)

    @_debug
    def start(self):
        Thread.start(self)


@dc.dataclass
class HasThread(ThreadBase):
    """This ThreadBase contains a thread, and is constructed with a callback"""

    callback: Callback = print
    daemon: bool = False
    error: t.Callable = none
    looping: bool = False
    name: str = ''

    def __post_init__(self):
        ThreadBase.__init__(self)

    @_debug(after=True)
    def join(self, timeout: t.Optional[float] = None):
        self.thread.join(timeout)

    @_debug
    def start(self):
        self.thread.start()

    def new_thread(self) -> Thread:
        return Thread(target=self.run, daemon=self.daemon)

    @cached_property
    def thread(self) -> Thread:
        return self.new_thread()



import typing as t
from warnings import warn

import ibis
import pandas
from ibis.backends.base import BaseBackend

from aladindb.backends.base.data_backend import BaseDataBackend
from aladindb.backends.ibis.db_helper import get_db_helper
from aladindb.backends.ibis.field_types import FieldType, dtype
from aladindb.backends.ibis.query import Table
from aladindb.backends.ibis.utils import get_output_table_name
from aladindb.backends.local.artifacts import FileSystemArtifactStore
from aladindb.backends.sqlalchemy.metadata import SQLAlchemyMetadata
from aladindb.components.model import APIModel, Model
from aladindb.components.schema import Schema

BASE64_PREFIX = 'base64:'


class IbisDataBackend(BaseDataBackend):
    def __init__(self, conn: BaseBackend, name: str, in_memory: bool = False):
        super().__init__(conn=conn, name=name)
        self.in_memory = in_memory
        self.dialect = getattr(conn, 'name', 'base')
        self.db_helper = get_db_helper(self.dialect)

    def url(self):
        return self.conn.con.url + self.name

    def build_artifact_store(self):
        return FileSystemArtifactStore(conn='.aladindb/artifacts/', name='ibis')

    def build_metadata(self):
        return SQLAlchemyMetadata(conn=self.conn.con, name='ibis')

    def create_ibis_table(self, identifier: str, schema: Schema):
        self.conn.create_table(identifier, schema=schema)

    def insert(self, table_name, raw_documents):
        for doc in raw_documents:
            for k, v in doc.items():
                doc[k] = self.db_helper.convert_data_format(v)
        table_name, raw_documents = self.db_helper.process_before_insert(
            table_name, raw_documents
        )
        if not self.in_memory:
            self.conn.insert(table_name, raw_documents)
        else:
            self.conn.create_table(table_name, pandas.DataFrame(raw_documents))

    def create_model_table_or_collection(self, model: t.Union[Model, APIModel]):
        msg = (
            "Model must have an encoder to create with the"
            f" {type(self).__name__} backend."
        )
        assert model.encoder is not None, msg
        if isinstance(model.encoder, FieldType):
            output_type = dtype(model.encoder.identifier)
        else:
            output_type = model.encoder
        fields = {
            'output_id': dtype('str'),  # TODO why do we need this????
            'input_id': dtype('str'),
            'query_id': dtype('string'),
            'output': output_type,
            'key': dtype('string'),
        }
        return Table(
            identifier=get_output_table_name(model.identifier, model.version),
            schema=Schema(
                identifier=f'_schema/{model.identifier}/{model.version}', fields=fields
            ),
        )

    def create_table_and_schema(self, identifier: str, mapping: dict):
        """
        Create a schema in the data-backend.
        """

        try:
            mapping = self.db_helper.process_schema_types(mapping)
            t = self.conn.create_table(identifier, schema=ibis.schema(mapping))
        except Exception as e:
            if 'exists' in str(e):
                warn("Table already exists, skipping...")
                t = self.conn.table(identifier)
            else:
                raise e
        return t

    def drop(self, force: bool = False):
        raise NotImplementedError(
            "Dropping tables needs to be done in each DB natively"
        )

    def get_table_or_collection(self, identifier):
        return self.conn.table(identifier)

    def disconnect(self):
        """
        Disconnect the client
        """

        # TODO: implement me



"""
The classes in this file define the configuration variables for aladindb,
which means that this file gets imported before alost anything else, and
canot contain any other imports from this project.
"""

import dataclasses as dc
import json
import os
import typing as t
from enum import Enum

_CONFIG_IMMUTABLE = True


@dc.dataclass
class BaseConfigJSONable:
    _lock: t.ClassVar[bool] = False

    def force_set(self, name, value):
        """
        Forcefully setattr of BaseConfigJSONable instance
        """
        super().__setattr__(name, value)

    def __setattr__(self, name, value):
        if not _CONFIG_IMMUTABLE or self._lock is False:
            super().__setattr__(name, value)
            return

        raise AttributeError(
            f'Process attempted to set "{name}" attribute of immutable configuration '
            f'object {self}.'
        )


@dc.dataclass
class Retry(BaseConfigJSONable):
    """
    Describes how to retry using the `tenacity` library

    :param stop_after_attempt: The number of attempts to make
    :param wait_max: The maximum time to wait between attempts
    :param wait_min: The minimum time to wait between attempts
    :param wait_multiplier: The multiplier for the wait time between attempts
    """

    stop_after_attempt: int = 2
    wait_max: float = 10.0
    wait_min: float = 4.0
    wait_multiplier: float = 1.0


@dc.dataclass
class Cluster(BaseConfigJSONable):
    """
    Describes a connection to distributed work via Dask

    :param backfill_batch_size: The number of rows to backfill at a time
                                for vector-search loading
    :param compute: The URI for compute i.e 'local', 'dask+tcp://localhost:8786'
                    "None": Run all jobs in local mode i.e simple function call
                    "local": same as above
                    "dask+thread": Run all jobs on a local threaded dask cluster
                    "dask+tcp://<host>:<port>": Run all jobs on a remote dask cluster

    :param vector_search: The URI for the vector search service
                          "None": Run vector search on local
                          "http://<host>:<port>": Connect a remote vector search service
    :param cdc: The URI for the change data capture service (if "None"
                then no cdc assumed)
                "None": Run cdc on local as a thread.
                "http://<host>:<port>": Connect a remote cdc service
    """

    compute: str = 'local'  # 'dask+tcp://local', 'dask+thread', 'local'
    vector_search: str = 'in_memory'  # '<in_memory|lance>://localhost:8000'
    cdc: t.Optional[str] = None  # 'http://localhost:8001'  # None
    backfill_batch_size: int = 100

    @property
    def vector_search_type(self):
        search_type = self.vector_search.split('://')[0]
        if search_type == 'http':
            # Return default vector_search
            return 'in_memory'
        return search_type

    @property
    def is_remote_vector_search(self):
        split = self.vector_search.split('://')
        dialect = split[0]
        return dialect != 'mongodb+srv' and len(split) > 1


class LogLevel(str, Enum):
    """
    Enumerate log severity level
    """

    DEBUG = 'DEBUG'
    INFO = 'INFO'
    SUCCESS = "SUCCESS"
    WARN = 'WARN'
    ERROR = 'ERROR'


class LogType(str, Enum):
    """
    Enumerate the standard logs
    """

    # SYSTEM uses the systems STDOUT and STDERR for printing the logs.
    # DEBUG, INFO, and WARN go to STDOUT.
    # ERROR goes to STDERR.
    SYSTEM = "SYSTEM"

    # LOKI a format that is compatible with the Loki Log aggregation system.
    LOKI = "LOKI"


class BytesEncoding(str, Enum):
    BYTES = 'Bytes'
    BASE64 = 'Str'


@dc.dataclass
class Config(BaseConfigJSONable):
    """
    The data class containing all configurable aladindb values

    :param data_backend: The URI for the data backend
    :param vector_search: The configuration for the vector search {'in_memory', 'lance'}
    :param artifact_store: The URI for the artifact store
    :param metadata_store: The URI for the metadata store
    :param cluster: Settings distributed computing and change data capture
    :param retries: Settings for retrying failed operations

    :param downloads_folder: Settings for downloading files

    :param fold_probability: The probability of validation fold

    :param log_level: The severity level of the logs
    :param logging_type: The type of logging to use

    :param bytes_encoding: The encoding of bytes in the data backend

    """

    data_backend: str = 'mongodb://aladin:aladin@localhost:27017/test_db'

    lance_home: str = os.path.join('.aladindb', 'vector_indices')

    artifact_store: t.Optional[str] = None
    metadata_store: t.Optional[str] = None

    cluster: Cluster = dc.field(default_factory=Cluster)
    retries: Retry = dc.field(default_factory=Retry)

    downloads_folder: t.Optional[str] = None
    fold_probability: float = 0.05

    log_level: LogLevel = LogLevel.INFO
    logging_type: LogType = LogType.SYSTEM

    dot_env: t.Optional[str] = None

    bytes_encoding: BytesEncoding = BytesEncoding.BYTES

    def __post_init__(self):
        if self.dot_env:
            import dotenv

            dotenv.load_dotenv(self.dot_env)
        self._lock = True

    @property
    def self_hosted_vector_search(self) -> bool:
        return self.data_backend == self.cluster.vector_search

    @property
    def hybrid_storage(self):
        return self.downloads_folder is not None

    @property
    def comparables(self):
        """
        A dict of `self` excluding some defined attributes.
        """
        _dict = dc.asdict(self)
        list(map(_dict.pop, ('cluster', 'retries', 'downloads_folder')))
        return _dict

    def dict(self):
        return dc.asdict(self)

    def match(self, cfg: t.Dict):
        """
        Match the target cfg dict with `self` comparables dict.
        """
        self_cfg = self.comparables

        self_hash = hash(json.dumps(self_cfg, sort_keys=True))
        cfg_hash = hash(json.dumps(cfg, sort_keys=True))
        return self_hash == cfg_hash

    def force_set(self, name, value):
        """
        Brings immutable behaviour to `CFG` instance.

        CAUTION: Only use it in development mode with caution,
        as this can bring unexpected behaviour.
        """
        parent = self
        names = name.split('.')
        if len(names) > 1:
            name = names[-1]
            for n in names[:-1]:
                parent = getattr(parent, n)
            parent.force_set(name, value)
        else:
            return super().force_set(name, value)



import hashlib
import typing as t

from aladindb import CFG


def get_file_from_uri(uri):
    """
    Get file name from uri.

    >>> _get_file('file://test.txt')
    'test.txt'
    >>> _get_file('http://test.txt')
    '414388bd5644669b8a92e45a96318890f6e8de54'
    """
    if uri.startswith('file://'):
        file = uri[7:]
    elif (
        uri.startswith('http://')
        or uri.startswith('https://')
        or uri.startswith('s3://')
    ):
        file = f'{CFG.downloads_folder}/{hashlib.sha1(uri.encode()).hexdigest()}'
    else:
        raise NotImplementedError(f'File type of {file} not supported')
    return file


def load_uris(
    r: dict,
    root: str,
    encoders: t.Dict,
    raises: bool = False,
):
    """
    Load ``"bytes"`` into ``"_content"`` from ``"uri"`` inside ``r``.

    :param r: The dict to load the bytes into
    :param root: The root directory to load the bytes from
    :param raises: Whether to raise an error if the file is not found

    >>> with open('/tmp/test.txt', 'wb') as f:
    ...     _ = f.write(bytes('test', 'utf-8'))
    >>> r = {"_content": {"uri": "file://test.txt"}}
    >>> load_uris(r, '/tmp')
    >>> r
    {'_content': {'uri': 'file://test.txt', 'bytes': b'test'}}
    """
    if not isinstance(r, dict):
        return
    for k, v in r.items():
        if isinstance(v, dict):
            if (
                k == '_content'
                and 'uri' in v
                and 'bytes' not in v
                and encoders[v['encoder']].load_hybrid
            ):
                file = get_file_from_uri(v['uri'])
                if root:
                    file = f'{root}/{file}'
                try:
                    with open(f'{file}', 'rb') as f:
                        r['_content']['bytes'] = f.read()
                except FileNotFoundError as e:
                    if raises:
                        raise e
            else:
                load_uris(v, root, encoders=encoders)
        elif isinstance(v, list):
            for x in v:
                load_uris(x, root, encoders=encoders)
        else:
            pass



import dataclasses as dc
import typing as t

import numpy as np
from overrides import override

from aladindb.base.datalayer import Datalayer
from aladindb.base.document import Document
from aladindb.components.component import Component
from aladindb.components.encoder import Encoder
from aladindb.components.listener import Listener
from aladindb.ext.utils import str_shape
from aladindb.misc.annotations import public_api
from aladindb.misc.special_dicts import MongoStyleDict
from aladindb.vector_search.base import VectorIndexMeasureType

if t.TYPE_CHECKING:
    pass


@public_api(stability='stable')
@dc.dataclass(kw_only=True)
class VectorIndex(Component):
    """
    A component carrying the information to apply a vector index to a ``DB`` instance
    {component_parameters}
    :param indexing_listener: Listener which is applied to created vectors
    :param compatible_listener: Listener which is applied to vectors to be compared
    :param measure: Measure to use for comparison
    :param metric_values: Metric values for this index
    """

    __doc__ = __doc__.format(component_parameters=Component.__doc__)

    type_id: t.ClassVar[str] = 'vector_index'

    indexing_listener: t.Union[Listener, str]
    compatible_listener: t.Union[None, Listener, str] = None
    measure: VectorIndexMeasureType = VectorIndexMeasureType.cosine
    metric_values: t.Optional[t.Dict] = dc.field(default_factory=dict)

    @override
    def on_load(self, db: Datalayer) -> None:
        if isinstance(self.indexing_listener, str):
            self.indexing_listener = t.cast(
                Listener, db.load('listener', self.indexing_listener)
            )

        if isinstance(self.compatible_listener, str):
            self.compatible_listener = t.cast(
                Listener, db.load('listener', self.compatible_listener)
            )

    @property
    def child_components(self) -> t.Sequence[t.Tuple[str, str]]:
        out = [('indexing_listener', 'listener')]
        if self.compatible_listener is not None:
            out.append(('compatible_listener', 'listener'))
        return out

    def get_vector(
        self,
        like: Document,
        models: t.List[str],
        keys: t.List[str],
        db: t.Any = None,
        outputs: t.Optional[t.Dict] = None,
    ):
        document = MongoStyleDict(like.unpack())
        if outputs is not None:
            outputs = outputs or {}
            if '_outputs' not in document:
                document['_outputs'] = {}
            document['_outputs'].update(outputs)

            assert not isinstance(self.indexing_listener, str)
        available_keys = list(document.keys()) + ['_base']
        try:
            model_name, key = next(
                (m, k) for m, k in zip(models, keys) if k in available_keys
            )
        except StopIteration:
            raise Exception(
                f'Keys in provided {like} don\'t match'
                f'VectorIndex keys: {keys}, with model: {models}'
            )

        model_input = document
        if key == '_base' and key in document:
            model_input = document[key]
        elif key != '_base':
            model_input = document[key]

        model = db.models[model_name]
        return (
            model.predict(model_input, one=True),
            model.identifier,
            key,
        )

    def get_nearest(
        self,
        like: Document,
        db: t.Any,
        id_field: str = '_id',
        outputs: t.Optional[t.Dict] = None,
        ids: t.Optional[t.Sequence[str]] = None,
        n: int = 100,
    ) -> t.Tuple[t.List[str], t.List[float]]:
        """Given a document, find the nearest results in this vector index, returned as
        two parallel lists of result IDs and scores

        :param like: The document to compare against
        :param db: The datastore to use
        :param outputs: An optional dictionary
        :param ids: A list of ids to match
        :param n: Number of items to return
        """

        models, keys = self.models_keys
        if len(models) != len(keys):
            raise ValueError(f'len(model={models}) != len(keys={keys})')
        within_ids = ids or ()

        if isinstance(like.content, dict) and id_field in like.content:
            return db.fast_vector_searchers[self.identifier].find_nearest_from_id(
                str(like[id_field]), within_ids=within_ids, limit=n
            )
        h = self.get_vector(
            like=like,
            models=models,
            keys=keys,
            db=db,
            outputs=outputs,
        )[0]

        return db.fast_vector_searchers[self.identifier].find_nearest_from_array(
            h, within_ids=within_ids, n=n
        )

    @property
    def models_keys(self) -> t.Tuple[t.List[str], t.List[str]]:
        """
        Return a list of model and keys for each listener
        """
        assert not isinstance(self.indexing_listener, str)
        assert not isinstance(self.compatible_listener, str)

        if self.compatible_listener:
            listeners = [self.indexing_listener, self.compatible_listener]
        else:
            listeners = [self.indexing_listener]

        models = [w.model.identifier for w in listeners]  # type: ignore[union-attr]
        keys = [w.key for w in listeners]
        return models, keys

    @property
    def dimensions(self) -> int:
        assert not isinstance(self.indexing_listener, str)
        assert not isinstance(self.indexing_listener.model, str)
        if shape := getattr(self.indexing_listener.model.encoder, 'shape', None):
            return shape[-1]
        raise ValueError('Couldn\'t get shape of model outputs from model encoder')


class EncodeArray:
    def __init__(self, dtype):
        self.dtype = dtype

    def __call__(self, x):
        x = np.asarray(x)
        if x.dtype != self.dtype:
            raise TypeError(f'dtype was {x.dtype}, expected {self.dtype}')
        return memoryview(x).tobytes()


class DecodeArray:
    def __init__(self, dtype):
        self.dtype = dtype

    def __call__(self, bytes):
        return np.frombuffer(bytes, dtype=self.dtype).tolist()


def vector(shape):
    """
    Create an encoder for a vector (list of ints/ floats) of a given shape

    :param shape: The shape of the vector
    """
    return Encoder(
        identifier=f'vector[{str_shape(shape)}]',
        shape=shape,
        encoder=None,
        decoder=None,
    )


def sqlvector(shape):
    """
    Create an encoder for a vector (list of ints/ floats) of a given shape
    compatible with sql databases.

    :param shape: The shape of the vector
    """
    return Encoder(
        identifier=f'sqlvector[{str_shape(shape)}]',
        shape=shape,
        encoder=EncodeArray(dtype='float64'),
        decoder=DecodeArray(dtype='float64'),
    )



def get_output_table_name(model_identifier, version):
    """Get the output table name for the given model."""
    # use `_` to connect the model_identifier and version
    return f'_outputs_{model_identifier}_{version}'



import dataclasses as dc
import time
import typing as t
from queue import Empty, Queue

from .runnable import Event


@dc.dataclass
class QueueChunker:
    """Chunk a queue into lists of length at most `chunk_size` within time `timeout`
    :param chunk_size: Maximum number of entries in a chunk
    :param timeout: Maximum amount of time to block
    :param accumulate_timeouts: If accumulate timeouts is True, then `timeout` is
                                the total timeout allowed over the whole chunk,
                                otherwise the timeout is applied to each item.
    """

    chunk_size: int
    timeout: float
    accumulate_timeouts: bool = False

    def __call__(self, queue: Queue, stop_event: Event) -> t.Iterator[t.List]:
        def chunk():
            start = self.accumulate_timeouts and time.time()

            for _ in range(self.chunk_size):
                if stop_event.is_set():
                    return

                elapsed = self.accumulate_timeouts and time.time() - start
                timeout = self.timeout - elapsed

                try:
                    item = queue.get(timeout=timeout)
                except Empty:
                    return
                else:
                    yield item

        while not stop_event:
            if c := list(chunk()):
                yield c



import json

from typer import Option

from aladindb import CFG

from . import command


@command(help='Print all the aladindb configs as JSON')
def config(
    schema: bool = Option(
        False, '--schema', '-s', help='If set, print the JSON schema for the model'
    ),
):
    d = CFG.schema() if schema else CFG.dict()
    print(json.dumps(d, indent=2))



import asyncio
import dataclasses as dc
from typing import Any, List, Optional

import aiohttp
import requests

from aladindb import logging
from aladindb.ext.llm.base import BaseLLMAPI, BaseLLMModel, BaseOpenAI

VLLM_INFERENCE_PARAMETERS_LIST = [
    "n",
    "best_of",
    "presence_penalty",
    "frequency_penalty",
    "repetition_penalty",
    "temperature",
    "top_p",
    "top_k",
    "min_p",
    "use_beam_search",
    "length_penalty",
    "early_stopping",
    "stop",
    "stop_token_ids",
    "include_stop_str_in_output",
    "ignore_eos",
    "max_tokens",
    "logprobs",
    "prompt_logprobs",
    "skip_special_tokens",
    "spaces_between_special_tokens",
    "logits_processors",
]


@dc.dataclass
class VllmOpenAI(BaseOpenAI):
    """
    Wrapper for requesting the vLLM API service
    (OpenAI format, started by vllm.entrypoints.openai.api_server)

    {parent_doc}
    """

    openai_api_key: Optional[str] = "EMPTY"

    __doc__ = __doc__.format(parent_doc=BaseOpenAI.__doc__)


@dc.dataclass
class VllmAPI(BaseLLMAPI):
    """
    Wrapper for requesting the vLLM API service
    (API Server format, started by vllm.entrypoints.api_server)
    {parent_doc}
    """

    __doc__ = __doc__.format(parent_doc=BaseLLMAPI.__doc__)

    def _generate(self, prompt: str, **kwargs) -> str:
        """
        Batch generate text from a prompt.
        """
        post_data = self.build_post_data(prompt, **kwargs)
        response = requests.post(self.api_url, json=post_data)
        return response.json()["text"][0]

    def _batch_generate(self, prompts: List[str], **kwargs: Any) -> List[str]:
        """
        Use asyncio to batch generate text from a list of prompts.
        """
        return asyncio.run(self._async_batch_generate(prompts, **kwargs))

    async def _async_generate(self, session, semaphore, prompt: str, **kwargs) -> str:
        post_data = self.build_post_data(prompt, **kwargs)
        async with semaphore:
            try:
                async with session.post(self.api_url, json=post_data) as response:
                    response_json = await response.json()
                    return response_json["text"][0]
            except aiohttp.ClientError as e:
                logging.error(f"HTTP request failed: {e}. Prompt: {prompt}")
                return ""

    async def _async_batch_generate(self, prompts: List[str], **kwargs) -> List[str]:
        async with aiohttp.ClientSession() as session:
            semaphore = asyncio.Semaphore(self.max_batch_size or len(prompts))
            tasks = [
                self._async_generate(session, semaphore, prompt, **kwargs)
                for prompt in prompts
            ]
            return await asyncio.gather(*tasks)

    def build_post_data(self, prompt: str, **kwargs: dict[str, Any]) -> dict[str, Any]:
        total_kwargs = {}
        for key, value in {**self.inference_kwargs, **kwargs}.items():
            if key in VLLM_INFERENCE_PARAMETERS_LIST:
                total_kwargs[key] = value
        return {"prompt": prompt, **total_kwargs}


@dc.dataclass
class VllmModel(BaseLLMModel):
    """
    Load a large language model from VLLM.

    :param model_name: The name of the model to use.
    :param trust_remote_code: Whether to trust remote code.
    :param dtype: The data type to use.
    {parent_doc}
    """

    __doc__ = __doc__.format(parent_doc=BaseLLMModel.__doc__)

    tensor_parallel_size: int = 1
    trust_remote_code: bool = True
    vllm_kwargs: dict = dc.field(default_factory=dict)

    def __post_init__(self):
        self.on_ray = self.on_ray or bool(self.ray_address)
        super().__post_init__()

    def init(self):
        try:
            from vllm import LLM
        except ImportError:
            raise Exception("You must install vllm with command 'pip install vllm'")

        if self.on_ray:
            try:
                import ray
            except ImportError:
                raise Exception("You must install vllm with command 'pip install ray'")

            runtime_env = {"pip": ["vllm"]}
            if not ray.is_initialized():
                ray.init(address=self.ray_address, runtime_env=runtime_env)

            LLM = ray.remote(LLM).remote

        self.llm = LLM(
            model=self.model_name,
            tensor_parallel_size=self.tensor_parallel_size,
            trust_remote_code=self.trust_remote_code,
            **self.vllm_kwargs,
        )

    def _batch_generate(self, prompts: List[str], **kwargs: Any) -> List[str]:
        from vllm import SamplingParams

        # support more parameters
        sampling_params = SamplingParams(
            **self.get_kwargs(SamplingParams, kwargs, self.inference_kwargs)
        )

        if self.on_ray:
            import ray

            results = ray.get(
                self.llm.generate.remote(prompts, sampling_params, use_tqdm=False)
            )
        else:
            results = self.llm.generate(prompts, sampling_params, use_tqdm=False)

        return [result.outputs[0].text for result in results]

    def _generate(self, prompt: str, **kwargs: Any) -> str:
        return self._batch_generate([prompt], **kwargs)[0]



import os
import typing as t

import lance
import numpy
import pyarrow as pa

from aladindb import CFG
from aladindb.vector_search.base import BaseVectorSearcher, VectorItem


class LanceVectorSearcher(BaseVectorSearcher):
    """
    Implementation of a vector index using the ``lance`` library.

    :param identifier: Unique string identifier of index
    :param dimensions: Dimension of the vector embeddings in the Lance dataset
    :param h: Seed vectors ``numpy.ndarray``
    :param index: list of IDs
    :param measure: measure to assess similarity
    """

    def __init__(
        self,
        identifier: str,
        dimensions: int,
        h: t.Optional[numpy.ndarray] = None,
        index: t.Optional[t.List[str]] = None,
        measure: t.Optional[str] = None,
    ):
        self.dataset_path = os.path.join(CFG.lance_home, f'{identifier}.lance')
        self.dimensions = dimensions
        self._created = False
        self.measure = measure
        if h is not None:
            self._create_or_append_to_dataset(h, index, mode='create')

    @property
    def dataset(self):
        if not os.path.exists(self.dataset_path):
            self._create_or_append_to_dataset([], [])
        return lance.dataset(self.dataset_path)

    def __len__(self):
        return self.dataset.count_rows()

    def _create_or_append_to_dataset(self, vectors, ids, mode: str = 'create'):
        if not self._created:
            if not os.path.exists(self.dataset_path):
                os.makedirs(self.dataset_path, exist_ok=True)
                mode = 'create'
            else:
                self._created = True
                mode = 'append'
        type = pa.list_(
            pa.field('values', pa.float32(), nullable=False), self.dimensions
        )
        vectors = self.to_list(vectors)
        _vecs = pa.array([v for v in vectors], type=type)
        _ids = pa.array(ids, type=pa.string())
        _table = pa.Table.from_arrays([_ids, _vecs], names=['id', 'vector'])
        lance.write_dataset(_table, self.dataset_path, mode=mode)
        self._created = True

    def add(self, items: t.Sequence[VectorItem]) -> None:
        ids = [item.id for item in items]
        vectors = [item.vector for item in items]
        self._create_or_append_to_dataset(vectors, ids, mode='append')

    def delete(self, ids: t.Sequence[str]) -> None:
        to_remove = ", ".join(f"'{str(id)}'" for id in ids)
        self.dataset.delete(f"id IN ({to_remove})")

    def find_nearest_from_id(
        self,
        _id,
        n: int = 100,
        within_ids: t.Sequence[str] = (),
    ) -> t.Tuple[t.List[str], t.List[float]]:
        # The ``lance`` file format has been specifically designed for fast
        # random access. The logic to take advantage of this is implemented
        # by the ``.take`` method.

        # See https://blog.lancedb.com/benchmarking-random-access-in-lance-ed690757a826
        # for further details.

        vector = self.dataset.take([int(_id)], columns=['vector']).to_pydict()[
            'vector'
        ][0]
        return self.find_nearest_from_array(vector, n=n, within_ids=within_ids)

    def find_nearest_from_array(
        self,
        h: numpy.typing.ArrayLike,
        n: int = 100,
        within_ids: t.Sequence[str] = (),
    ) -> t.Tuple[t.List[str], t.List[float]]:
        # NOTE: filter is currently applied AFTER vector-search
        # See https://lancedb.github.io/lance/api/python/lance.html#lance.dataset.LanceDataset.scanner
        if within_ids:
            assert (
                type(within_ids) == tuple
            ), 'within_ids must be a tuple for lance sql parser'
            result = self.dataset.to_table(
                columns=['id'],
                nearest={
                    'column': 'vector',
                    'q': h,
                    'k': n,
                    'metric': self.measure,
                },
                filter=f"id in {within_ids}",
                offset=0,
            )
        else:
            result = self.dataset.to_table(
                columns=['id'],
                nearest={"column": 'vector', "q": h, "k": n},
                offset=0,
            )
        ids = result['id'].to_pylist()
        scores = result['_distance'].to_pylist()
        return ids, scores



from typing import List, Optional

import aiohttp
import requests
from aiohttp import ClientConnectionError, ClientResponseError
from requests.exceptions import HTTPError

from aladindb.ext.utils import get_key
from aladindb.misc.retry import Retry

JINA_API_URL: str = "https://api.jina.ai/v1/embeddings"
KEY_NAME = 'JINA_API_KEY'

retry = Retry(exception_types=(ClientResponseError, ClientConnectionError, HTTPError))


class JinaAPIClient:
    def __init__(
        self,
        api_key: Optional[str] = None,
        model_name: str = 'jina-embeddings-v2-base-en',
    ):
        """
        Create a JinaAPIClient to provide an interface to encode using
            Jina Embedding platform sync and async.

        :param api_key: The Jina API key.
            It can be explicitly provided or automatically read
            from the environment variable JINA_API_KEY (recommended).
        :param model_name: The name of the Jina model to use.
         Check the list of available models on `https://jina.ai/embeddings/`
        """
        # if the user does not provide the API key,
        # check if it is set in the environment variable
        if api_key is None:
            api_key = get_key(KEY_NAME)

        self.model_name = model_name
        self._session = requests.Session()
        self._headers = {
            "Authorization": f"Bearer {api_key}",
            "Accept-Encoding": "identity",
            "Content-type": "application/json",
        }
        self._session.headers.update(self._headers)

    @retry
    def encode_batch(self, texts: List[str]) -> List[List[float]]:
        response = self._session.post(
            JINA_API_URL, json={"input": texts, "model": self.model_name}
        ).json()
        if "data" not in response:
            raise RuntimeError(response["detail"])

        # Sort resulting embeddings by index
        sorted_embeddings = sorted(response["data"], key=lambda e: e["index"])
        embeddings = [result["embedding"] for result in sorted_embeddings]
        return embeddings

    @retry
    async def aencode_batch(self, texts: List[str]) -> List[List[float]]:
        async with aiohttp.ClientSession() as session:
            payload = {
                'model': self.model_name,
                'input': texts,
            }

            async with session.post(
                JINA_API_URL,
                headers=self._headers,
                json=payload,
            ) as response:
                response.raise_for_status()
                response_json = await response.json()
                if "data" not in response_json:
                    raise RuntimeError(response_json["detail"])

                # Sort resulting embeddings by index
                sorted_embeddings = sorted(
                    response_json["data"], key=lambda e: e["index"]
                )
                embeddings = [result["embedding"] for result in sorted_embeddings]
                return embeddings



import typing as t

import numpy

from aladindb.components.encoder import Encoder
from aladindb.ext.utils import str_shape


class EncodeArray:
    def __init__(self, dtype):
        self.dtype = dtype

    def __call__(self, x):
        if x.dtype != self.dtype:
            raise TypeError(f'dtype was {x.dtype}, expected {self.dtype}')
        return memoryview(x).tobytes()


class DecodeArray:
    def __init__(self, dtype):
        self.dtype = dtype

    def __call__(self, bytes):
        return numpy.frombuffer(bytes, dtype=self.dtype)


def array(dtype: str, shape: t.Sequence):
    """
    Create an encoder of numpy arrays.

    :param dtype: The dtype of the array.
    :param shape: The shape of the array.
    """
    return Encoder(
        identifier=f'numpy.{dtype}[{str_shape(shape)}]',
        encoder=EncodeArray(dtype),
        decoder=DecodeArray(dtype),
        shape=shape,
    )



import dataclasses as dc
import typing as t

from overrides import override

from aladindb import CFG
from aladindb.backends.base.query import CompoundSelect
from aladindb.base.datalayer import Datalayer
from aladindb.base.document import _OUTPUTS_KEY
from aladindb.misc.annotations import public_api
from aladindb.misc.server import request_server

from ..jobs.job import Job
from .component import Component
from .model import Model


@public_api(stability='stable')
@dc.dataclass(kw_only=True)
class Listener(Component):
    """
    Listener object which is used to process a column/ key of a collection or table,
    and store the outputs.
    {component_parameters}
    :param key: Key to be bound to model
    :param model: Model for processing data
    :param select: Object for selecting which data is processed
    :param identifier: A string used to identify the model.
    :param active: Toggle to ``False`` to deactivate change data triggering
    :param predict_kwargs: Keyword arguments to self.model.predict
    """

    __doc__ = __doc__.format(component_parameters=Component.__doc__)

    key: str
    model: t.Union[str, Model]
    select: CompoundSelect
    identifier: t.Optional[str] = None  # type: ignore[assignment]
    active: bool = True
    predict_kwargs: t.Optional[t.Dict] = dc.field(default_factory=dict)

    type_id: t.ClassVar[str] = 'listener'

    def __post_init__(self):
        if self.identifier is None and self.model is not None:
            if isinstance(self.model, str):
                self.identifier = f'{self.model}/{self.id_key}'
            else:
                self.identifier = f'{self.model.identifier}/{self.id_key}'
        super().__post_init__()

    @property
    def outputs(self):
        return f'{_OUTPUTS_KEY}.{self.key}.{self.model.identifier}.{self.model.version}'

    @property
    def child_components(self) -> t.Sequence[t.Tuple[str, str]]:
        """Returns a list of child components as pairs TBD"""
        return [('model', 'model')]

    @override
    def pre_create(self, db: Datalayer) -> None:
        if isinstance(self.model, str):
            self.model = t.cast(Model, db.load('model', self.model))

        if self.select is not None and self.select.variables:
            self.select = t.cast(CompoundSelect, self.select.set_variables(db))

    def post_create(self, db: Datalayer) -> None:
        # Start cdc service if enabled
        if self.select is not None and self.active and not db.server_mode:
            if CFG.cluster.cdc:
                request_server(
                    service='cdc',
                    endpoint='listener/add',
                    args={'name': self.identifier},
                    type='get',
                )
            else:
                db.cdc.add(self)

    @property
    def dependencies(self) -> t.List[str]:
        out = []
        if self.key.startswith('_outputs.'):
            _, key, model = self.key.split('.')
            out.append(f'{model}/{key}')
        if self.select.output_fields:
            out.extend([f'{v}/{k}' for k, v in self.select.output_fields.items()])
        return out

    @property
    def id_key(self) -> str:
        if self.key.startswith('_outputs.'):
            return self.key.split('.')[1]
        return self.key

    @override
    def schedule_jobs(
        self,
        db: Datalayer,
        dependencies: t.Sequence[Job] = (),
        verbose: bool = False,
    ) -> t.Sequence[t.Any]:
        """
        Schedule jobs for the listener

        :param database: The DB instance to process
        :param dependencies: A list of dependencies
        :param verbose: Whether to print verbose output
        """
        if not self.active:
            return []
        assert not isinstance(self.model, str)

        out = [
            self.model.predict(
                X=self.key,
                db=db,
                select=self.select,
                dependencies=dependencies,
                **(self.predict_kwargs or {}),
            )
        ]
        return out

    def cleanup(self, database: Datalayer) -> None:
        """Clean up when the listener is deleted

        :param database: The DB instance to process
        """
        # TODO - this doesn't seem to do anything
        if (cleanup := getattr(self.select, 'model_cleanup', None)) is not None:
            assert not isinstance(self.model, str)
            cleanup(database, model=self.model.identifier, key=self.key)



import subprocess
import typing as t
from subprocess import PIPE, CalledProcessError

__all__ = (
    'CalledProcessError',
    'PIPE',
    'run',
    'out',
)


def run(
    args: t.Sequence[str],
    text: bool = True,
    check: bool = True,
    verbose: bool = False,
    **kwargs,
) -> subprocess.CompletedProcess:
    """
    Run a command, printing it if verbose is enabled.

    :param args: The command to run.
    :param text: Whether to use text mode.
    :param check: Whether to raise an error if the command fails.
    :param **kwargs: Additional arguments to pass to ``subprocess.run``.
    """
    if verbose:
        print('$', *args)
    return subprocess.run(args, text=text, check=check, **kwargs)


def out(args: t.Sequence[str], **kwargs) -> str:
    """
    Run a command and return the output.

    :param args: The command to run.
    :param **kwargs: Additional arguments to pass to ``subprocess.run``.
    """
    return run(args, stdout=PIPE, **kwargs).stdout.strip()



import dataclasses as dc
import io
import pickle
import typing as t

import dill

from aladindb.base.artifact import Artifact
from aladindb.components.component import Component
from aladindb.misc.annotations import public_api

if t.TYPE_CHECKING:
    from aladindb.base.datalayer import Datalayer


@public_api(stability='beta')
@dc.dataclass(kw_only=True)
class Serializer(Component):
    """
    A component carrying the information to apply a serializer to a
    model.
    {component_parameters}
    :param object: The serializer
    """

    __doc__ = __doc__.format(component_parameters=Component.__doc__)

    type_id: t.ClassVar[str] = 'serializer'

    encoder: t.Union[t.Callable, Artifact]
    decoder: t.Union[t.Callable, Artifact]
    constructor: t.Union[t.Callable, Artifact, str, None] = None

    info: t.Optional[t.Dict] = None

    def __post_init__(self):
        super().__post_init__()

        if not isinstance(self.encoder, Artifact):
            self.encoder = Artifact(self.encoder)
        if not isinstance(self.decoder, Artifact):
            self.decoder = Artifact(self.decoder)
        if self.constructor and not isinstance(self.constructor, Artifact):
            self.constructor = Artifact(self.constructor)

    def pre_create(self, db: 'Datalayer'):
        super().pre_create(db)
        self.object = t.cast(t.Type, self.identifier)

    def encode(self, object: t.Any) -> bytes:
        assert isinstance(self.encoder, Artifact)
        return self.encoder.artifact(object, info=self.info)

    def decode(self, object: t.Any) -> bytes:
        assert isinstance(self.decoder, Artifact)
        return self.decoder.artifact(object, info=self.info)

    def construct(self):
        assert isinstance(self.constructor, Artifact)
        return self.constructor.artifact(**self.info)


def pickle_encode(object: t.Any, info: t.Optional[t.Dict] = None) -> bytes:
    return pickle.dumps(object)


def pickle_decode(b: bytes, info: t.Optional[t.Dict] = None) -> t.Any:
    return pickle.loads(b)


def dill_encode(object: t.Any, info: t.Optional[t.Dict] = None) -> bytes:
    return dill.dumps(object)


def dill_decode(b: bytes, info: t.Optional[t.Dict] = None) -> t.Any:
    return dill.loads(b)


def torch_encode(object: t.Any, info: t.Optional[t.Dict] = None) -> bytes:
    import torch

    from aladindb.ext.torch.utils import device_of

    if not isinstance(object, dict):
        previous_device = str(device_of(object))
        object.to('cpu')
        f = io.BytesIO()
        torch.save(object, f)
        object.to(previous_device)
    else:
        f = io.BytesIO()
        torch.save(object, f)
    return f.getvalue()


def torch_decode(b: bytes, info: t.Optional[t.Dict] = None) -> t.Any:
    import torch

    return torch.load(io.BytesIO(b))


class Serializers:
    serializers: t.Dict[str, Serializer] = {}

    def __iter__(self):
        return iter(self.serializers)

    def add(self, name: str, serializer: Serializer):
        self.serializers[name] = serializer

    def __getitem__(self, serializer):
        return self.serializers[serializer]


serializers = Serializers()
serializers.add(
    'pickle', Serializer('pickle', encoder=pickle_encode, decoder=pickle_decode)
)
serializers.add('dill', Serializer('dill', encoder=dill_encode, decoder=dill_decode))
serializers.add(
    'torch', Serializer('torch', encoder=torch_encode, decoder=torch_decode)
)


def encode_torch_state_dict(module, info):
    import torch

    buffer = io.BytesIO()
    torch.save(module.state_dict(), buffer)

    return buffer.getvalue()


class DecodeTorchStateDict:
    def __init__(self, cls):
        self.cls = cls

    def __call__(self, b: bytes, info: t.Dict):
        import torch

        buffer = io.BytesIO(b)
        module = self.cls(**info)
        module.load_state_dict(torch.load(buffer))
        return module


def build_torch_state_serializer(module, info):
    return Serializer(
        identifier=module.__name__,
        info=info,
        encoder=encode_torch_state_dict,
        decoder=DecodeTorchStateDict(module),
    )



# ruff: noqa: E402
from .base import config, configs, jsonable, logger
from .base.aladin import aladin

ICON = '🔮'
CFG = configs.CFG
JSONable = jsonable.JSONable
ROOT = configs.ROOT

logging = logger.Logging

__version__ = '0.1.1'

from aladindb.backends import ibis, mongodb

from .base.document import Document
from .components.dataset import Dataset
from .components.encoder import Encoder
from .components.listener import Listener
from .components.metric import Metric
from .components.model import Model
from .components.schema import Schema
from .components.serializer import Serializer
from .components.vector_index import VectorIndex, vector

__all__ = (
    'CFG',
    'ICON',
    'JSONable',
    'ROOT',
    'config',
    'logging',
    'aladin',
    'Encoder',
    'Document',
    'Model',
    'Listener',
    'VectorIndex',
    'vector',
    'Dataset',
    'Metric',
    'Schema',
    'Serializer',
    'mongodb',
    'ibis',
)



from __future__ import annotations

import dataclasses as dc
import typing as t
from functools import wraps

import networkx
from networkx import DiGraph, ancestors

from .job import ComponentJob, FunctionJob, Job

if t.TYPE_CHECKING:
    from aladindb.base.datalayer import Datalayer


@dc.dataclass
class TaskWorkflow:
    """
    Keep a graph of jobs that need to be performed and their dependencies,
    and perform them when called.

    :param database: ``DB`` instance to use
    :param G: ``networkx.DiGraph`` to use as the graph
    """

    database: Datalayer
    G: DiGraph = dc.field(default_factory=DiGraph)

    @wraps(DiGraph.add_edge)
    def add_edge(self, node1: str, node2: str) -> None:
        self.G.add_edge(node1, node2)

    @wraps(DiGraph.add_node)
    def add_node(self, node: str, job: t.Union[FunctionJob, ComponentJob]) -> None:
        self.G.add_node(node, job=job)

    def watch(self) -> None:
        """Watch the stdout of each job in this workflow in topological order"""
        for node in list(networkx.topological_sort(self.G)):
            self.G.nodes[node]['job'].watch()

    def run_jobs(
        self,
    ):
        """Run all the jobs in this workflow"""
        pred = self.G.predecessors
        current_group = [n for n in self.G.nodes if not ancestors(self.G, n)]
        done = set()

        while current_group:
            for node in current_group:
                job: Job = self.G.nodes[node]['job']
                dependencies = [self.G.nodes[a]['job'].future for a in pred(node)]
                job(
                    self.database,
                    dependencies=dependencies,
                )
                done.add(node)

            current_group = [
                n for n in self.G.nodes if set(pred(n)) <= done and n not in done
            ]



import dataclasses as dc
import importlib
import inspect
import typing as t

from aladindb.misc.serialization import asdict


def is_component_metadata(r: dict) -> bool:
    COMPONENT_KEYS = {'type_id', 'identifier', 'version'}
    if COMPONENT_KEYS == set(r):
        return True
    return False


def is_component(r: dict) -> bool:
    COMPONENT_KEYS = {'cls', 'dict', 'module'}
    if COMPONENT_KEYS <= set(r):
        return True
    return False


def _deserialize(r: t.Any, db: None = None) -> t.Any:
    if isinstance(r, (list, tuple)):
        return [_deserialize(i, db=db) for i in r]

    if not isinstance(r, dict):
        return r

    if not is_component(r):
        return {k: _deserialize(v, db=db) for k, v in r.items()}

    module = importlib.import_module(r['module'])
    component_cls = getattr(module, r['cls'])

    kwargs = _deserialize(r['dict'])
    if 'db' in inspect.signature(component_cls.__init__).parameters:
        kwargs.update(db=db)

    instance = component_cls(**{k: v for k, v in kwargs.items() if k != 'version'})

    # special handling of Component.version
    from aladindb.components.component import Component

    if issubclass(component_cls, Component):
        instance.version = r.get('version', None)

    return instance


def _serialize(item: t.Any, serialize_variables: bool = True):
    if isinstance(item, Serializable):
        sub_dict = {
            k: _serialize(getattr(item, k), serialize_variables=serialize_variables)
            for k in item.dict()
        }
        out = {
            'cls': item.__class__.__name__,
            'dict': sub_dict,
            'module': item.__class__.__module__,
        }
        from aladindb.components.component import Component

        if isinstance(item, Component):
            out.update(
                {
                    'type_id': item.type_id,
                    'identifier': item.identifier,
                    'version': getattr(
                        item,
                        'version',
                        None,
                    ),  # type: ignore[dict-item]
                }
            )
        return out
    if isinstance(item, dict):
        return {
            k: _serialize(v, serialize_variables=serialize_variables)
            for k, v in item.items()
        }
    if isinstance(item, (list, tuple)):
        return [_serialize(x, serialize_variables=serialize_variables) for x in item]
    if serialize_variables:
        if isinstance(item, Variable):
            return {
                'cls': 'Variable',
                'dict': {'value': item.value, 'setter_callback': item.setter_callback},
                'module': 'aladindb.base.serializable',
            }
    return item


class VariableError(Exception):
    ...


def _find_variables(r):
    from .document import Document

    if isinstance(r, dict):
        return sum([_find_variables(v) for v in r.values()], [])
    elif isinstance(r, (list, tuple)):
        return sum([_find_variables(v) for v in r], [])
    elif isinstance(r, Variable):
        return [r]
    elif isinstance(r, Document):
        return r.variables
    return []


def _replace_variables(x, db, **kwargs):
    from .document import Document

    if isinstance(x, dict):
        return {
            _replace_variables(k, db, **kwargs): _replace_variables(v, db, **kwargs)
            for k, v in x.items()
        }
    if isinstance(x, (list, tuple)):
        return [_replace_variables(v, db) for v in x]
    if isinstance(x, Variable):
        return x.set(db, **kwargs)
    if isinstance(x, Document):
        return x.set_variables(db, **kwargs)
    return x


@dc.dataclass
class Serializable:
    """
    Base class for serializable objects. This class is used to serialize and
    deserialize objects to and from JSON + Artifact instances.
    """

    deserialize = staticmethod(_deserialize)
    serialize = _serialize

    @property
    def variables(self) -> t.List['Variable']:
        out = {}
        v = _find_variables(self.dict())
        for var in v:
            out[var.value] = var
        return sorted(list(out.values()), key=lambda x: x.value)

    def set_variables(self, db, **kwargs) -> 'Serializable':
        r = self.serialize(
            serialize_variables=False
        )  # get serialized version of class instance
        r = _replace_variables(r, db, **kwargs)  # replace variables with values
        return self.deserialize(r)  # rebuild new class instance

    def dict(self) -> t.Dict[str, t.Any]:
        return asdict(self)


class Variable:
    """
    Mechanism for allowing "free variables" in a serializable object.
    The idea is to allow a variable to be set at runtime, rather than
    at object creation time.

    :param value: The name of the variable to be set at runtime.
    :param setter_callback: A callback function that takes the value, datalayer
                            and kwargs as input and returns the formatted
                            variable.
    """

    def __init__(
        self, value: t.Union[str, t.Any], setter_callback: t.Optional[t.Callable] = None
    ):
        self.value = value
        self.setter_callback = setter_callback

    def __repr__(self) -> str:
        return '$' + str(self.value)

    def __hash__(self) -> int:
        return hash(self.value)

    def set(self, db, **kwargs):
        """
        Get the intended value from the values of the global variables.

        >>> Variable('number').set(db, number=1.5, other='test')
        1.5

        :param db: The datalayer instance.
        :param kwargs: Variables to be used in the setter_callback
                       or as formatting variables.
        """
        if self.setter_callback is not None:
            try:
                return self.setter_callback(db, self.value, kwargs)
            except Exception as e:
                raise VariableError(
                    f'Could not set variable {self.value} '
                    f'based on {self.setter_callback} and **kwargs: {kwargs}'
                ) from e
        else:
            assert isinstance(self.value, str)
            return kwargs[self.value]



import base64
import dataclasses as dc
import io
import pickle
import typing as t

from aladindb import CFG
from aladindb.base.artifact import Artifact
from aladindb.base.config import BytesEncoding
from aladindb.components.component import Component
from aladindb.misc.annotations import public_api

Decode = t.Callable[[bytes], t.Any]
Encode = t.Callable[[t.Any], bytes]


def _pickle_decoder(b: bytes) -> t.Any:
    return pickle.load(io.BytesIO(b))


def _pickle_encoder(x: t.Any) -> bytes:
    f = io.BytesIO()
    pickle.dump(x, f)
    return f.getvalue()


@public_api(stability='stable')
@dc.dataclass(kw_only=True)
class Encoder(Component):
    """
    Storeable ``Component`` allowing byte encoding of primary data,
    i.e. data inserted using ``db.base.db.Datalayer.insert``
    {component_parameters}
    :param identifier: Unique identifier
    :param decoder: callable converting a ``bytes`` string to a ``Encodable``
                    of this ``Encoder``
    :param encoder: Callable converting an ``Encodable`` of this ``Encoder``
                    to ``bytes``
    :param shape: Shape of the data
    :param load_hybrid: Whether to load the data from the URI or return the URI in
                        `CFG.hybrid` mode
    """

    __doc__ = __doc__.format(component_parameters=Component.__doc__)

    type_id: t.ClassVar[str] = 'encoder'

    artifact_artibutes: t.ClassVar[t.Sequence[str]] = ['decoder', 'encoder']
    decoder: t.Union[t.Callable, Artifact] = dc.field(
        default_factory=lambda: Artifact(artifact=_pickle_decoder)
    )
    encoder: t.Union[t.Callable, Artifact] = dc.field(
        default_factory=lambda: Artifact(artifact=_pickle_encoder)
    )
    shape: t.Optional[t.Sequence] = None
    load_hybrid: bool = True

    # TODO what's this for?
    encoders: t.ClassVar[t.List] = []

    def __post_init__(self):
        super().__post_init__()

        self.encoders.append(self.identifier)
        if isinstance(self.decoder, t.Callable):
            self.decoder = Artifact(artifact=self.decoder)
        if isinstance(self.encoder, t.Callable):
            self.encoder = Artifact(artifact=self.encoder)

    def __call__(
        self, x: t.Optional[t.Any] = None, uri: t.Optional[str] = None
    ) -> 'Encodable':
        return Encodable(self, x=x, uri=uri)

    def decode(
        self, b: t.Union[bytes, str], bytes_encoding: t.Optional[BytesEncoding] = None
    ) -> t.Any:
        assert isinstance(self.decoder, Artifact)
        bytes_encoding = bytes_encoding or CFG.bytes_encoding

        if (
            CFG.bytes_encoding == BytesEncoding.BASE64
            or bytes_encoding == BytesEncoding.BASE64
        ):
            b = self.from_base64(b)
        return self(self.decoder.artifact(b))

    def dump(self, other):
        return self.encoder.artifact(other)

    @staticmethod
    def to_base64(bytes):
        return base64.b64encode(bytes).decode('utf-8')

    @staticmethod
    def from_base64(encoded):
        return base64.b64decode(encoded)

    def encode(
        self,
        x: t.Optional[t.Any] = None,
        uri: t.Optional[str] = None,
        wrap: bool = True,
        bytes_encoding: t.Optional[BytesEncoding] = None,
    ) -> t.Union[t.Optional[str], t.Dict[str, t.Any]]:
        # TODO clarify what is going on here

        def _encode(x):
            bytes_ = self.encoder.artifact(x)
            if (
                CFG.bytes_encoding == BytesEncoding.BASE64
                or bytes_encoding == BytesEncoding.BASE64
            ):
                bytes_ = self.to_base64(bytes_)
            return bytes_

        def _wrap_content(x):
            return {
                '_content': {
                    'bytes': _encode(x),
                    'encoder': self.identifier,
                }
            }

        if self.encoder is not None:
            if x is not None:
                if wrap:
                    return _wrap_content(x)
                return _encode(x)
            else:
                if wrap:
                    return {
                        '_content': {
                            'uri': uri,
                            'encoder': self.identifier,
                        }
                    }
                return uri
        else:
            assert x is not None
            return x


@dc.dataclass
class Encodable:
    """
    Data variable wrapping encode-able item. Encoding is controlled by the referred
    to ``Encoder`` instance.

    :param encoder: Instance of ``Encoder`` controlling encoding
    :param x: Wrapped content
    :param uri: URI of the content, if any
    """

    encoder: Encoder
    x: t.Optional[t.Any] = None
    uri: t.Optional[str] = None

    def encode(
        self, bytes_encoding: t.Optional[BytesEncoding] = None
    ) -> t.Union[t.Optional[str], t.Dict[str, t.Any]]:
        bytes_encoding = bytes_encoding or CFG.bytes_encoding
        assert hasattr(self.encoder, 'encode')
        return self.encoder.encode(
            x=self.x, uri=self.uri, bytes_encoding=bytes_encoding
        )


default_encoder = Encoder(identifier='_default')



from ibis.backends.base import BaseBackend
from pymongo import MongoClient

from aladindb.backends.ibis.data_backend import IbisDataBackend
from aladindb.backends.local.artifacts import FileSystemArtifactStore
from aladindb.backends.mongodb.artifacts import MongoArtifactStore
from aladindb.backends.mongodb.data_backend import MongoDataBackend
from aladindb.backends.mongodb.metadata import MongoMetaDataStore
from aladindb.backends.sqlalchemy.metadata import SQLAlchemyMetadata
from aladindb.vector_search.atlas import MongoAtlasVectorSearcher
from aladindb.vector_search.in_memory import InMemoryVectorSearcher
from aladindb.vector_search.lance import LanceVectorSearcher

data_backends = {
    'mongodb': MongoDataBackend,
    'ibis': IbisDataBackend,
}

artifact_stores = {
    'mongodb': MongoArtifactStore,
    'filesystem': FileSystemArtifactStore,
}

metadata_stores = {
    'mongodb': MongoMetaDataStore,
    'sqlalchemy': SQLAlchemyMetadata,
}

vector_searcher_implementations = {
    'lance': LanceVectorSearcher,
    'in_memory': InMemoryVectorSearcher,
    'mongodb+srv': MongoAtlasVectorSearcher,
}

CONNECTIONS = {
    'pymongo': MongoClient,
    'ibis': BaseBackend,
}



import math
import typing as t

from fastapi import Request

from aladindb.base.datalayer import Datalayer
from aladindb.ext.utils import aladindecode
from aladindb.vector_search.base import VectorItem

ListVectorType = t.Union[t.List[t.Union[float, int]], t.Dict]

VectorSearchResultType = t.Tuple[t.List[str], t.List[float]]


def _vector_search(
    x: t.Union[str, ListVectorType],
    n: int,
    vector_index: str,
    by_array: bool = True,
    db=None,
) -> VectorSearchResultType:
    vi = db.fast_vector_searchers[vector_index]
    if by_array:
        x = aladindecode(x, db.encoders)
        ids, scores = vi.searcher.find_nearest_from_array(x, n=n)
    else:
        ids, scores = vi.searcher.find_nearest_from_id(x, n=n)
    scores = [-1.0 if math.isnan(s) else s for s in scores]
    return ids, scores


def database(request: Request) -> Datalayer:
    return request.app.state.pool


def list_search(db: Datalayer):
    '''
    Helper functon for listing all vector search indices.
    :param db: Datalayer instance
    '''
    return db.show('vector_index')


def create_search(vector_index: str, db=None):
    '''
    Initiates a vector search class corresponding to `vector_index`
    :param vector_index: Vector class to initiate
    :param db: Datalayer instance
    '''
    db.fast_vector_searchers.update(
        {vector_index: db.initialize_vector_searcher(vector_index, backfill=True)}
    )


def query_search_from_array(
    array: ListVectorType, vector_index: str, n: int = 100, db=None
) -> VectorSearchResultType:
    '''
    Perform a vector search with an array
    :param array: Array to perform vector search on index.
    :param vector_index: Vector search index
    :param db: Datalayer instance
    '''
    return _vector_search(array, n=n, vector_index=vector_index, db=db)


def query_search_from_id(
    id: str, vector_index: str, n: int = 100, db=None
) -> VectorSearchResultType:
    '''
    Perform a vector search with an id
    :param id: Identifier for vector
    :param vector_index: Vector search index
    :param db: Datalayer instance
    '''
    return _vector_search(id, n=n, vector_index=vector_index, db=db, by_array=False)


def add_search(vector, vector_index: str, db=None):
    '''
    Adds a vector in vector index `vector_index`
    :param vector: Vector to be added.
    :param vector_index: Vector index where vector needs to be added.

    '''
    vector = [VectorItem(id=v.id, vector=v.vector) for v in vector]

    vi = db.fast_vector_searchers[vector_index]
    vi.searcher.add(vector)


def delete_search(ids: t.List[str], vector_index: str, db=None):
    '''
    Deletes a vector corresponding to `id`
    '''
    vi = db.fast_vector_searchers[vector_index]
    vi.searcher.delete(ids)



import copy
import json
import re
import typing as t
from functools import cached_property

import pymongo

from aladindb import CFG, logging
from aladindb.vector_search.base import BaseVectorSearcher

if t.TYPE_CHECKING:
    from aladindb.components.vector_index import VectorIndex


class MongoAtlasVectorSearcher(BaseVectorSearcher):
    """
    Implementation of atlas vector search

    :param identifier: Unique string identifier of index
    """

    def __init__(
        self,
        identifier: str,
        collection: str,
        dimensions: t.Optional[int] = None,
        measure: t.Optional[str] = None,
        output_path: t.Optional[str] = None,
        **kwargs,
    ):
        self.identifier = identifier
        self.database: pymongo.database.Database = pymongo.MongoClient(
            CFG.cluster.vector_search
        ).get_database()
        assert output_path
        self.output_path = output_path
        self.collection = collection
        self.measure = measure
        self.dimensions = dimensions

        if not self._check_if_exists(identifier):
            self._create_index(collection, output_path)

    def __len__(self):
        pass

    @cached_property
    def index(self):
        return self.database[self.collection]

    @classmethod
    def from_component(cls, vi: 'VectorIndex'):
        from aladindb.components.listener import Listener
        from aladindb.components.model import Model

        assert isinstance(vi.indexing_listener, Listener)
        collection = vi.indexing_listener.select.table_or_collection.identifier

        indexing_key = vi.indexing_listener.key
        if indexing_key.startswith('_outputs'):
            indexing_key = indexing_key.split('.')[1]
        assert isinstance(vi.indexing_listener.model, Model)
        assert isinstance(collection, str), 'Collection is required to be a string'
        indexing_model = vi.indexing_listener.model.identifier
        indexing_version = vi.indexing_listener.model.version
        output_path = f'_outputs.{indexing_key}.{indexing_model}.{indexing_version}'

        return MongoAtlasVectorSearcher(
            identifier=vi.identifier,
            dimensions=vi.dimensions,
            measure=vi.measure,
            output_path=output_path,
            collection=collection,
        )

    def _replace_document_with_vector(self, step):
        step = copy.deepcopy(step)
        assert "like" in step['$vectorSearch']
        vector = step['$vectorSearch']['like']
        step['$vectorSearch']['queryVector'] = vector

        step['$vectorSearch']['path'] = self.output_path
        step['$vectorSearch']['index'] = self.identifier
        del step['$vectorSearch']['like']
        return step

    def _prepare_pipeline(self, pipeline):
        pipeline = copy.deepcopy(pipeline)
        try:
            search_step = next(
                (i, step) for i, step in enumerate(pipeline) if '$vectorSearch' in step
            )
        except StopIteration:
            return pipeline
        pipeline[search_step[0]] = self._replace_document_with_vector(
            search_step[1],
        )
        return pipeline

    def _find(self, h, n=100):
        h = self.to_list(h)
        pl = [
            {
                "$vectorSearch": {
                    'like': h,
                    "limit": n,
                    'numCandidates': n,
                }
            },
            {'$addFields': {'score': {'$meta': 'vectorSearchScore'}}},
        ]
        pl = self._prepare_pipeline(
            pl,
        )
        cursor = self.index.aggregate(pl)
        scores = []
        ids = []
        for vector in cursor:
            scores.append(vector['score'])
            ids.append(str(vector['_id']))
        return ids, scores

    def find_nearest_from_id(self, id: str, n=100, within_ids=None):
        h = self.index.find_one({'id': id})
        return self.find_nearest_from_array(h, n=n, within_ids=within_ids)

    def find_nearest_from_array(self, h, n=100, within_ids=None):
        return self._find(h, n=n)

    def add(self, items):
        items = list(map(lambda x: x.to_dict(), items))
        if not CFG.cluster.vector_search == CFG.data_backend:
            self.index.insert_many(items)

    def delete(self, items):
        ids = list(map(lambda x: x.id, items))
        if not CFG.cluster.vector_search == CFG.data_backend:
            self.index.delete_many({'id': {'$in': ids}})

    def _create_index(self, collection: str, output_path: str):
        """
        Create a vector index in the data backend if an Atlas deployment.

        :param vector_index: vector index to create
        """
        _, key, model, version = output_path.split('.')
        if re.match('^_outputs\.[A-Za-z0-9_]+\.[A-Za-z0-9_]+', key):
            key = key.split('.')[1]
        fields4 = {
            str(version): [
                {
                    "dimensions": self.dimensions,
                    "similarity": self.measure,
                    "type": "knnVector",
                }
            ]
        }
        fields3 = {
            model: {
                "fields": fields4,
                "type": "document",
            }
        }
        fields2 = {
            key: {
                "fields": fields3,
                "type": "document",
            }
        }
        fields1 = {
            "_outputs": {
                "fields": fields2,
                "type": "document",
            }
        }
        index_definition = {
            "createSearchIndexes": collection,
            "indexes": [
                {
                    "name": self.identifier,
                    "definition": {
                        "mappings": {
                            "dynamic": True,
                            "fields": fields1,
                        }
                    },
                }
            ],
        }
        logging.info(json.dumps(index_definition, indent=2))
        self.database.command(index_definition)

    def _check_if_exists(self, index: str):
        indexes = self.index.list_search_indexes()
        return len(
            [i for i in indexes if i['name'] == index and i['status'] == 'READY']
        )



import sys
import warnings
from typing import Optional


def public_api(stability: str = 'stable'):
    """Annotation for documenting public APIs.

    If ``stability="alpha"``, the API can be used by advanced users who are
    tolerant to and expect breaking changes.

    If ``stability="beta"``, the API is still public and can be used by early
    users, but are subject to change.

    If ``stability="stable"``, the APIs will remain backwards compatible across
    minor releases.
    """

    assert stability in ["stable", "beta", "alpha"]

    def wrap(obj):
        if stability in ["alpha", "beta"]:
            message = (
                f"**public_api({stability}):** This API is in {stability} "
                "and may change before becoming stable."
            )
            _append_doc(obj, message=message)
        return obj

    return wrap


class aladindbDeprecationWarning(DeprecationWarning):
    """Specialized Deprecation Warning for fine grained filtering control"""

    pass


if not sys.warnoptions:
    warnings.filterwarnings("module", category=aladindbDeprecationWarning)


def _append_doc(obj, *, message: str, directive: Optional[str] = None):
    if not obj.__doc__:
        obj.__doc__ = ""

    obj.__doc__ = obj.__doc__.rstrip()

    indent = _get_indent(obj.__doc__)
    obj.__doc__ += "\n\n"

    if directive is not None:
        obj.__doc__ += f"{' ' * indent}.. {directive}::\n\n"

        message = message.replace("\n", "\n" + " " * (indent + 4))
        obj.__doc__ += f"{' ' * (indent + 4)}{message}"
    else:
        message = message.replace("\n", "\n" + " " * (indent + 4))
        obj.__doc__ += f"{' ' * indent}{message}"
    obj.__doc__ += f"\n{' ' * indent}"


def _get_indent(docstring: str) -> int:
    if not docstring:
        return 0

    non_empty_lines = list(filter(bool, docstring.splitlines()))
    if len(non_empty_lines) == 1:
        return 0

    return len(non_empty_lines[1]) - len(non_empty_lines[1].lstrip())



from aladindb import logging


class ComponentInUseError(Exception):
    pass


class ComponentInUseWarning(Warning):
    pass


class BaseException(Exception):
    '''
    BaseException which logs a message after
    exception
    '''

    def __init__(self, msg):
        self.msg = msg
        logging.exception(self.msg, e=self)

    def __str__(self):
        return self.msg


class ServiceRequestException(BaseException):
    '''
    ServiceRequestException
    '''


class QueryException(BaseException):
    '''
    QueryException
    '''


class DatabackendException(BaseException):
    '''
    DatabackendException
    '''


class MetadataException(BaseException):
    '''
    MetadataException
    '''


class ComponentException(BaseException):
    '''
    ComponentException
    '''



import hashlib
import typing as t

import typing_extensions as te

if t.TYPE_CHECKING:
    from aladindb.backends.base.artifact import ArtifactStore

ArtifactCache = t.Dict[int, t.Any]


class Artifact:
    """
    An artifact from a computation that can be serialized or deserialized

    :param artifact: the computed artifact, which may be of any type
    :param file_id: a key used to identify the file in the ``ArtifactStore``
    :param info: the ``info`` dictionary is passed into
                 ``ArtifactStore.create_artifact()``
    :param object_id: the Python ``id()`` of the artifact
    :param save_method: the name of the finalization method on the artifact
                        to call before saving
    :param serializer: the name of the serializer
    :param sha1: the sha1 hash of the artifact
    :param hash: in case the object isn't hashable (deduplication not possible)
    """

    # TODO Why do we need all this?? The annotations are clear below
    file_id: t.Optional[str] = None
    object_id: int = 0
    save_method: t.Optional[str] = None
    serializer: str = 'dill'
    hash: t.Optional[int] = None
    uri: t.Optional[str] = None

    def __init__(
        self,
        artifact: t.Any = None,
        file_id: t.Optional[str] = None,
        object_id: int = 0,
        serializer: str = 'dill',
        sha1: str = '',
        hash: t.Optional[int] = None,
        artifact_store: t.Optional['ArtifactStore'] = None,
    ):
        self._artifact = artifact
        self.file_id = file_id
        self.object_id = object_id
        self.serializer = serializer
        self.hash = hash
        self._sha1 = sha1
        self.artifact_store = artifact_store

    @property
    def artifact(self):
        if self._artifact is None:
            msg = 'Artifact Store is not available.'
            assert self.artifact_store is not None, msg
            self._artifact = self.artifact_store.load_artifact(
                file_id=self.file_id,
                serializer=self.serializer,
            )
        return self._artifact

    @artifact.setter
    def artifact(self, value):
        self._artifact = value

    def load(self, artifact_store):
        self._artifact = artifact_store.load_artifact(
            file_id=self.file_id,
            serializer=self.serializer,
        )

    def sha1(self, serializers):
        if not self._sha1:
            b = self.serialize(serializers)
            self._sha1 = hashlib.sha1(b).hexdigest()
            return self._sha1
        return self._sha1

    def __hash__(self):
        if self.hash is not None:
            return self.hash

        if isinstance(self.artifact, list):
            return hash(str(self.artifact[:100]))
        if isinstance(self.artifact, dict):
            return hash(str(self.artifact))
        else:
            return hash(self.artifact)

    def __eq__(self, other):
        return self.artifact == other.artifact

    def __repr__(self):
        if self._artifact:
            return (
                f'<Artifact artifact={str(self.artifact)} serializer={self.serializer}>'
            )
        else:
            return f'<Artifact artifact={self.file_id} serializer={self.serializer}>'

    @staticmethod
    def _is_self_serializable(object):
        if 'serialize' in dir(object) and 'deserialize' in dir(object):
            return True
        return False

    def serialize(self, serializers) -> bytes:
        """Serialize this artifact into bytes"""

        if self._is_self_serializable(self.artifact):
            assert hasattr(object, 'serialize')
            return t.cast(bytes, object.serialize())

        serializer = serializers[self.serializer]
        try:
            return serializer.encode(self.artifact)
        except Exception as e:
            raise ArtifactSavingError(
                f'Error using serializer "{self.serializer}" '
                f'to save artifact: {self.artifact}'
            ) from e

    def save(self, artifact_store: 'ArtifactStore') -> t.Dict[str, t.Any]:
        """Store this artifact, and return a dictionary of the results

        :param artifact_store: the store to save the Artifact in
        """
        b = self.serialize(artifact_store.serializers)
        file_id = artifact_store.create(bytes=b)
        return {
            'file_id': file_id,
            'sha1': self.sha1(artifact_store.serializers),
            'serializer': self.serializer,
        }


class ArtifactDesc(te.TypedDict):
    """A description of an artifact in an artifact store

    :param file_id: A string identifying the artifact in the artifact store
    :param serializer: The name of the serializer used for the artifact store
    """

    file_id: str
    serializer: str


class ArtifactSavingError(Exception):
    pass



import typing as t

from pydantic import BaseModel

from aladindb import CFG
from aladindb.base.datalayer import Datalayer
from aladindb.server.app import DatalayerDependency, aladinApp
from aladindb.vector_search.server import service

assert (
    CFG.cluster.is_remote_vector_search
), "Set a correct uri for `cluster.vector_search`"


port = int(CFG.cluster.vector_search.split(':')[-1])
app = aladinApp('vector_search', port=port)


class VectorItem(BaseModel):
    id: str
    vector: service.ListVectorType


@app.add("/create/search", status_code=200, method='get')
def create_search(vector_index: str, db: Datalayer = DatalayerDependency()):
    service.create_search(vector_index=vector_index, db=db)
    return {'message': 'Vector index created successfully'}


@app.add("/query/id/search", method='post')
def query_search_by_id(
    id: str, vector_index: str, n: int = 100, db: Datalayer = DatalayerDependency()
):
    ids, scores = service.query_search_from_id(
        id, vector_index=vector_index, n=n, db=db
    )
    return {'ids': ids, 'scores': scores}


@app.add("/query/search")
def query_search_by_array(
    vector: service.ListVectorType,
    vector_index: str,
    n: int = 100,
    db: Datalayer = DatalayerDependency(),
):
    ids, scores = service.query_search_from_array(
        vector, vector_index=vector_index, n=n, db=db
    )
    return {'ids': ids, 'scores': scores}


@app.add("/add/search")
def add_search(
    vectors: t.List[VectorItem],
    vector_index: str,
    db: Datalayer = DatalayerDependency(),
):
    service.add_search(vectors, vector_index=vector_index, db=db)
    return {'message': 'Added vectors successfully'}


@app.add("/delete/search")
def delete_search(
    ids: t.List[str], vector_index: str, db: Datalayer = DatalayerDependency()
):
    service.delete_search(ids, vector_index=vector_index, db=db)
    return {'message': 'Ids deleted successfully'}


@app.add("/list/search")
def list_search(db: Datalayer = DatalayerDependency()):
    return service.list_search(db)



import dataclasses as dc
import os
import typing as t
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path

import yaml

from . import config_dicts
from .config import Config

File = t.Union[Path, str]

# The top-level directory of the project
ROOT = Path(__file__).parents[2]

# The default prefix used for config environment variables
PREFIX = 'aladindb_'

# The name of the environment variable used to read the config files.
# This value needs to be read before all the other config values are.
FILES_NAME = 'CONFIG_FILES'

# The base name of the configs file
CONFIG_FILE = 'configs.json'

_LOCAL_CONFIG = Path(CONFIG_FILE)
_PROJECT_CONFIG = ROOT / CONFIG_FILE
_USER_CONFIG = Path(f'~/.aladindb/{CONFIG_FILE}').expanduser()

_ALL_CONFIGS = _PROJECT_CONFIG, _LOCAL_CONFIG, _USER_CONFIG

FILE_SEP = ','


@dataclass(frozen=True)
class ConfigSettings:
    """
    A class that reads a Pydantic class from config files and environment variables.

    :param cls: The Pydantic class to read.
    :param default_files: The default config files to read.
    :param prefix: The prefix to use for environment variables.
    :param environ: The environment variables to read from.
    """

    cls: t.Type
    default_files: t.Union[t.Sequence[Path], str]
    prefix: str
    environ: t.Optional[t.Dict] = None
    base_config: t.Optional[Config] = None

    @cached_property
    def config(self) -> t.Any:
        """Read a Pydantic class"""

        if isinstance(self.base_config, dict):
            parent = self.base_config
        elif isinstance(self.base_config, Config):
            parent = self.base_config.dict()
        else:
            parent = self.cls().dict()

        env = dict(os.environ if self.environ is None else self.environ)
        env = config_dicts.environ_to_config_dict('aladindb_', parent, env)

        config_path = '.aladindb/config.yaml'
        if os.path.exists(config_path):
            with open(config_path) as f:
                kwargs = yaml.safe_load(f)
        else:
            kwargs = {}

        kwargs = config_dicts.combine_configs((parent, kwargs, env))

        def _dataclass_from_dict(cls, d):
            try:
                fieldtypes = {f.name: f.type for f in dc.fields(cls)}
                return cls(**{f: _dataclass_from_dict(fieldtypes[f], d[f]) for f in d})
            except Exception:
                return d

        return _dataclass_from_dict(self.cls, kwargs)


def build_config(cfg: t.Optional[Config] = None) -> Config:
    """
    Build the config object from the environment variables and config files.
    """
    CONFIG = ConfigSettings(Config, _ALL_CONFIGS, PREFIX, base_config=cfg)
    return CONFIG.config


CFG = build_config()



import uvicorn


def _aladin_local_service(name, app, cfg=None):
    from aladindb.server.app import Server

    service = getattr(cfg.cluster, name)
    assert isinstance(service, str)
    port = int(service.split(':')[-1])

    app.pre_start(cfg)
    config = uvicorn.Config(app.app, port=port, log_level="info")
    return Server(config=config)


class _Client:
    def __init__(self, db, services=()):
        self.db = db
        self.services = services
        for service in services:
            service.run_in_thread()

    def close(self):
        self.db.compute.shutdown()
        for service in self.services:
            service.stop()


def local_cluster(db):
    """
    This method is used to create a local cluster consisting of
    Vector search service, cdc service and dask setup.

    Once this cluster is up, user can offload vector search,
    cdc on these services.
    """
    from aladindb import CFG

    # Vector search local service
    CFG.force_set('cluster.vector_search', 'http://localhost:8000')
    CFG.force_set('cluster.cdc', 'http://localhost:8001')
    CFG.force_set('cluster.compute', 'dask+thread')

    from aladindb.vector_search.server.app import app as vector_search_app

    vector_search_server = _aladin_local_service(
        'vector_search', vector_search_app, CFG
    )

    # Cdc local service
    from aladindb.cdc.app import app as cdc_app

    cdc_server = _aladin_local_service('cdc', cdc_app, CFG)

    # Local compute
    from aladindb.backends.dask.compute import DaskComputeBackend

    local_compute = DaskComputeBackend('', local=True)
    db.set_compute(local_compute)

    return _Client(db=db, services=(cdc_server, vector_search_server))



import typing as t
from abc import ABC, abstractmethod, abstractproperty


class ComputeBackend(ABC):
    """
    Abstraction for sending jobs to a distributed compute platform.
    """

    @abstractproperty
    def type(self) -> str:
        """
        Return the type of compute engine
        """
        pass

    @abstractproperty
    def name(self) -> str:
        """
        Return the name of current compute engine
        """
        pass

    def get_local_client(self):
        '''Returns a local version of self'''
        pass

    @abstractmethod
    def submit(self, function: t.Callable, **kwargs) -> t.Any:
        """
        Submits a function to the Dask server for execution.

        :param function: The function to be executed.
        :param kwargs: Additional keyword arguments to be passed to the function.
        """
        pass

    @abstractproperty
    def tasks(self) -> t.Any:
        """
        List for all tasks
        """
        pass

    @abstractmethod
    def wait_all(self) -> None:
        """
        Waits for all pending tasks to complete.
        """
        pass

    @abstractmethod
    def result(self, identifier: str) -> t.Any:
        """
        Retrieves the result of a previously submitted task.
        Note: This will block until the future is completed.

        :param identifier: The identifier of the submitted task.
        """
        pass

    @abstractmethod
    def disconnect(self) -> None:
        """
        Disconnect the client.
        """
        pass

    @abstractmethod
    def shutdown(self) -> None:
        """
        Shuts down the compute cluster.
        """
        pass



import dataclasses as dc
import enum
import json
import re
import types
import typing as t

import pandas

from aladindb import Document, logging
from aladindb.backends.base.query import (
    CompoundSelect,
    Insert,
    Like,
    QueryComponent,
    QueryLinker,
    RawQuery,
    Select,
    TableOrCollection,
    _ReprMixin,
)
from aladindb.backends.ibis.cursor import aladinIbisResult
from aladindb.backends.ibis.utils import get_output_table_name
from aladindb.base.serializable import Variable
from aladindb.components.component import Component
from aladindb.components.encoder import Encoder
from aladindb.components.schema import Schema

if t.TYPE_CHECKING:
    from aladindb.base.datalayer import Datalayer

PRIMARY_ID: str = 'id'

IbisTableType = t.TypeVar('IbisTableType')
ParentType = t.TypeVar('ParentType')


class IbisBackendError(Exception):
    """
    This error represents ibis query related errors
    i.e when there is an error while executing an ibis query,
    use this exception to represent the error.
    """


@dc.dataclass(repr=False)
class IbisCompoundSelect(CompoundSelect):
    """
    A query incorporating vector-search and a standard ``ibis`` query
    """

    __doc__ = __doc__ + CompoundSelect.__doc__  # type: ignore[operator]

    def __hash__(self) -> int:
        return hash(json.dumps(self.serialize()))

    def __eq__(self, __value: object) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__eq__(__value)

    def __lt__(self, __value: object) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__lt__(__value)

    def __gt__(self, __value: object) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__gt__(__value)

    def __and__(self, __value: object) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__and__(__value)

    def __or__(self, __value: object) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__or__(__value)

    def __not__(self) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__not__()

    def __getitem__(self, item) -> bool:
        assert self.query_linker is not None
        return self.query_linker.__getitem__(item)

    def _get_query_linker(cls, table_or_collection, members) -> 'IbisQueryLinker':
        return IbisQueryLinker(table_or_collection=table_or_collection, members=members)

    @property
    def output_fields(self):
        return self.query_linker.output_fields

    def _get_query_component(
        self,
        name: str,
        type: str,
        args: t.Optional[t.Sequence] = None,
        kwargs: t.Optional[t.Dict] = None,
    ):
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        return IbisQueryComponent(name, type=type, args=args, kwargs=kwargs)

    def outputs(self, **kwargs):
        """
        This method returns a query which joins a query with the outputs
        for a table.

        :param key: The key on which the model was evaluated
        :param model: The model identifier for which to get the outputs
        :param version: The version of the model for which to get the outputs (optional)

        >>> q = t.filter(t.age > 25).outputs('txt', 'model_name')
        """
        assert self.query_linker is not None
        return IbisCompoundSelect(
            table_or_collection=self.table_or_collection,
            pre_like=self.pre_like,
            query_linker=self.query_linker._outputs(query_id=str(hash(self)), **kwargs),
            post_like=self.post_like,
        )

    def compile(self, db: 'Datalayer', tables: t.Optional[t.Dict] = None):
        """
        Convert the current query to an ``ibis`` native query.

        :param db: The aladindb connection
        :param tables: A dictionary of ``ibis`` tables to use for the query
        """
        assert self.pre_like is None, "pre_like must be None"
        assert self.post_like is None, "post_like must be None"
        assert self.query_linker is not None, "query_linker must be set"

        table_id = self.table_or_collection.identifier
        if tables is None:
            tables = {}
        if table_id not in tables:
            tables[table_id] = db.databackend.conn.table(table_id)
        return self.query_linker.compile(db, tables=tables)

    def get_all_tables(self):
        tables = [self.table_or_collection.identifier]
        if self.query_linker is not None:
            tables.extend(self.query_linker.get_all_tables())
        tables = list(set(tables))
        return tables

    def _get_all_fields(self, db):
        tables = self.get_all_tables()
        component_tables = []
        for tab in tables:
            component_tables.append(db.load('table', tab))
        fields = {}

        for tab in component_tables:
            fields_copy = tab.schema.fields.copy()
            if '_outputs' in tab.identifier and self.renamings:
                match = re.search(r"_outputs_(.*?)_(\d+)", tab.identifier)
                for k in self.renamings.values():
                    if match:
                        fields_copy[k] = fields_copy['output']
                        del fields_copy['output']
            else:
                for k in fields_copy:
                    if k in self.renamings.values():
                        fields_copy[k] = fields_copy[self.renamings[k]]
                        del fields_copy[k]
            fields.update(fields_copy)
        return fields

    @property
    def select_table(self):
        return self.table_or_collection

    def _execute_with_pre_like(self, db):
        assert self.pre_like is not None
        assert self.post_like is None
        similar_ids, similar_scores = self.pre_like.execute(db)
        similar_scores = dict(zip(similar_ids, similar_scores))

        query_linker_stub = self.table_or_collection.filter(
            getattr(self.table_or_collection, self.table_or_collection.primary_id).isin(
                similar_ids
            )
        )
        new_query_linker = query_linker_stub
        if self.query_linker is not None:
            new_query_linker = IbisQueryLinker(
                table_or_collection=self.table_or_collection,
                members=[
                    *query_linker_stub.members,
                    *self.query_linker.members,
                ],
            )

        return new_query_linker.execute(db), similar_scores

    def _execute_with_post_like(self, db):
        assert self.pre_like is None
        df = self.query_linker.select_ids.execute(db)
        query_ids = [id[0] for id in df.values.tolist()]
        similar_ids, similar_scores = self.post_like.execute(db, ids=query_ids)
        similar_scores = dict(zip(similar_ids, similar_scores))
        post_query_linker = self.query_linker.select_using_ids(similar_ids)
        return post_query_linker.execute(db), similar_scores

    def _execute(self, db):
        if self.pre_like is not None:
            return self._execute_with_pre_like(db)

        elif self.post_like is not None:
            return self._execute_with_post_like(db)

        return self.query_linker.execute(db), None

    @property
    def renamings(self):
        if self.query_linker is not None:
            return self.query_linker.renamings
        return {}

    def execute(self, db):
        output, scores = self._execute(db)
        fields = self._get_all_fields(db)

        for column in output.columns:
            try:
                type = fields[column]
            except KeyError:
                logging.warn(f'Disambiguation not yet supported of {column}: TODO!')
                continue

            if isinstance(type, Encoder):
                output[column] = output[column].map(type.decode)

        if scores is not None:
            output['scores'] = output[self.primary_id].map(scores)

        output = output.to_dict(orient='records')
        return aladinIbisResult(
            output, id_field=self.table_or_collection.primary_id, scores=scores
        )

    def select_ids_of_missing_outputs(self, key: str, model: str, version: int):
        """
        Query which selects ids where outputs are missing.
        """

        assert self.pre_like is None
        assert self.post_like is None
        assert self.query_linker is not None
        query_id = str(hash(self))

        out = self._query_from_parts(
            table_or_collection=self.table_or_collection,
            query_linker=self.query_linker._select_ids_of_missing_outputs(
                key=key,
                model=model,
                query_id=query_id,
                version=version,
            ),
        )
        return out

    def model_update(  # type: ignore[override]
        self,
        db,
        ids: t.List[t.Any],
        key: str,
        model: str,
        version: int,
        outputs: t.Sequence[t.Any],
        flatten: bool = False,
    ):
        if flatten:
            raise NotImplementedError('Flatten not yet supported for ibis')

        if key.startswith('_outputs'):
            key = key.split('.')[1]

        if not outputs:
            return

        query_id = str(hash(self))
        table_records = []
        for ix in range(len(outputs)):
            d = {
                'output_id': str(ids[ix]),
                'input_id': str(ids[ix]),
                'query_id': query_id,
                'output': outputs[ix],
                'key': key,
            }
            table_records.append(d)

        for r in table_records:
            if isinstance(r['output'], dict) and '_content' in r['output']:
                r['output'] = r['output']['_content']['bytes']
        db.databackend.insert(get_output_table_name(model, version), table_records)

    def add_fold(self, fold: str) -> Select:
        if self.query_linker is not None:
            # make sure we have a fold column in the query
            query_members = [
                i
                for i in self.query_linker.members
                if isinstance(i, IbisQueryComponent)
            ]
            if query_members:
                last_member = query_members[-1]
                if '_fold' not in last_member.args:
                    last_member.args = tuple(list(last_member.args) + ['_fold'])
        return self.filter(self._fold == fold)


class _LogicalExprMixin:
    def _logical_expr(self, members, collection, k, other: t.Optional[t.Any] = None):
        if other:
            args = [other]
        else:
            args = []
        members.append(
            IbisQueryComponent(k, args=args, kwargs={}, type=QueryType.QUERY)
        )
        return IbisQueryLinker(collection, members=members)

    def eq(self, other, members, collection):
        k = '__eq__'
        return self._logical_expr(members, collection, k, other=other)

    def or_(self, other, members, collection):
        k = '__or__'
        return self._logical_expr(members, collection, k, other=other)

    def not_(self, members, collection):
        k = '__not__'
        return self._logical_expr(members, collection, k)

    def and_(self, other, members, collection):
        k = '__and__'
        return self._logical_expr(members, collection, k, other=other)

    def gt(self, other, members, collection):
        k = '__gt__'
        return self._logical_expr(members, collection, k, other=other)

    def lt(self, other, members, collection):
        k = '__lt__'
        return self._logical_expr(members, collection, k, other=other)

    def getitem(self, other, members, collection):
        k = '__getitem__'
        return self._logical_expr(members[:], collection, k, other=other)


@dc.dataclass(repr=False)
class IbisQueryLinker(QueryLinker, _LogicalExprMixin):
    @property
    def renamings(self):
        out = {}
        for m in self.members:
            out.update(m.renamings)
        return out

    def __post_init__(self):
        self._output_fields = {}

    def repr_(self) -> str:
        out = super().repr_()
        out = re.sub('\. ', ' ', out)
        out = re.sub('\.\[', '[', out)
        return out

    @property
    def output_fields(self):
        return self._output_fields

    @output_fields.setter
    def output_fields(self, value):
        self._output_fields = value

    def __eq__(self, other):
        return self.eq(other, members=self.members, collection=self.table_or_collection)

    def __lt__(self, other):
        return self.lt(other, members=self.members, collection=self.table_or_collection)

    def __gt__(self, other):
        return self.gt(other, members=self.members, collection=self.table_or_collection)

    def __or__(self, other):
        return self.or_(
            other, members=self.members, collection=self.table_or_collection
        )

    def __and__(self, other):
        return self.and_(
            other, members=self.members, collection=self.table_or_collection
        )

    def __not__(self, other):
        return self.not_(
            other, members=self.members, collection=self.table_or_collection
        )

    def __getitem__(self, other):
        return self.getitem(
            other, members=self.members, collection=self.table_or_collection
        )

    @classmethod
    def _get_query_linker(cls, table_or_collection, members):
        return cls(
            table_or_collection=table_or_collection,
            members=members,
        )

    def _get_query_component(self, k):
        return IbisQueryComponent(name=k, type=QueryType.ATTR)

    @property
    def select_ids(self):
        return self.select(self.table_or_collection.primary_id)

    def select_single_id(self, id):
        return self.filter(
            self.table_or_collection.__getattr__(self.table_or_collection.primary_id)
            == id
        )

    def select_using_ids(self, ids):
        return self.filter(
            self.table_or_collection.__getattr__(
                self.table_or_collection.primary_id
            ).isin(ids)
        )

    def _select_ids_of_missing_outputs(
        self, key: str, model: str, query_id: str, version: int
    ):
        output_table = IbisQueryTable(
            identifier=get_output_table_name(model, version),
            primary_id='output_id',
        )
        filtered = output_table.filter(
            output_table.key == key and output_table.query_id == query_id
        )
        out = self.anti_join(
            filtered, filtered.input_id == self[self.table_or_collection.primary_id]
        )
        return out

    def get_all_tables(self):
        out = []
        for member in self.members:
            out.extend(member.get_all_tables())
        return list(set(out))

    def _outputs(self, query_id: str, **kwargs):
        for key, model in kwargs.items():
            version = None
            if '/' in model:
                model, version = model.split('/')
            symbol_table = IbisQueryTable(
                identifier=(
                    get_output_table_name(model, version)
                    if version is not None
                    else Variable(
                        get_output_table_name(model, '{version}'),
                        lambda db, value, kwargs: value.format(
                            version=db.show('model', model)[-1]
                        ),
                    )
                ),
                primary_id='output_id',
            )
            symbol_table = symbol_table.relabel(
                {
                    'output': (
                        f'_outputs.{key}.{model}.{version}'
                        if version is not None
                        else Variable(
                            f'_outputs.{key}.{model}' + '.{version}',
                            lambda db, value, kwargs: value.format(
                                version=db.show('model', model)[-1]
                            ),
                        )
                    )
                }
            )
            attr = getattr(  # type: ignore[call-overload]
                self, self.table_or_collection.primary_id
            )
            other_query = self.join(symbol_table, symbol_table.input_id == attr)
            other_query = other_query.filter(other_query.key == key)
            return other_query

    def compile(self, db: 'Datalayer', tables: t.Optional[t.Dict] = None):
        table_id = self.table_or_collection.identifier
        if tables is None:
            tables = {}
        if table_id not in tables:
            tables = {table_id: db.databackend.conn.table(table_id)}
        result = tables[table_id]
        for member in self.members:
            result, tables = member.compile(parent=result, db=db, tables=tables)
        return result, tables

    def execute(self, db):
        native_query, _ = self.compile(db)
        try:
            result = native_query.execute()
        except Exception as exc:
            raise IbisBackendError(
                f'{native_query} Wrong query or not supported yet :: {exc}'
            )
        for column in result.columns:
            result[column] = result[column].map(
                db.databackend.db_helper.recover_data_format
            )
        return result


class QueryType(str, enum.Enum):
    QUERY = 'query'
    ATTR = 'attr'


@dc.dataclass(repr=False, kw_only=True)
class Table(Component):
    """
    This is a representation of an SQL table in ibis,
    saving the important meta-data associated with the table
    in the ``aladindb`` meta-data store.
    {component_params}:param schema: The schema of the table
    :param primary_id: The primary id of the table
    """

    type_id: t.ClassVar[str] = 'table'
    __doc__ = __doc__.format(component_params=Component.__doc__)

    schema: Schema
    primary_id: str = 'id'

    def pre_create(self, db: 'Datalayer'):
        assert self.schema is not None, "Schema must be set"
        for e in self.schema.encoders:
            db.add(e)
        if db.databackend.in_memory:
            logging.info(f'Using in-memory tables "{self.identifier}" so doing nothing')
            return

        try:
            db.databackend.create_table_and_schema(self.identifier, self.schema.raw)
        except Exception as e:
            if 'already exists' in str(e):
                pass
            else:
                raise e

    @property
    def table_or_collection(self):
        return IbisQueryTable(self.identifier, primary_id=self.primary_id)

    def compile(self, db: 'Datalayer', tables: t.Optional[t.Dict] = None):
        return IbisQueryTable(self.identifier, primary_id=self.primary_id).compile(
            db, tables=tables
        )

    def insert(self, documents, **kwargs):
        return IbisQueryTable(
            identifier=self.identifier, primary_id=self.primary_id
        ).insert(documents, **kwargs)

    def like(self, r: 'Document', vector_index: str, n: int = 10):
        return IbisQueryTable(
            identifier=self.identifier, primary_id=self.primary_id
        ).like(r=r, vector_index=vector_index, n=n)

    def outputs(self, **kwargs):
        return IbisQueryTable(
            identifier=self.identifier, primary_id=self.primary_id
        ).outputs(**kwargs)

    def __getattr__(self, item):
        return getattr(
            IbisQueryTable(identifier=self.identifier, primary_id=self.primary_id), item
        )

    def __getitem__(self, item):
        return IbisQueryTable(
            identifier=self.identifier, primary_id=self.primary_id
        ).__getitem__(item)

    def to_query(self):
        return IbisCompoundSelect(
            table_or_collection=IbisQueryTable(
                self.identifier, primary_id=self.primary_id
            ),
            query_linker=IbisQueryLinker(
                table_or_collection=IbisQueryTable(
                    self.identifier, primary_id=self.primary_id
                )
            ),
        )


@dc.dataclass(repr=False)
class IbisQueryTable(_ReprMixin, TableOrCollection, Select):
    """
    This is a symbolic representation of a table
    for building ``IbisCompoundSelect`` queries.

    :param primary_id: The primary id of the table
    """

    primary_id: str = 'id'

    def compile(self, db: 'Datalayer', tables: t.Optional[t.Dict] = None):
        if tables is None:
            tables = {}
        if self.identifier not in tables:
            tables[self.identifier] = db.databackend.conn.table(self.identifier)
        return tables[self.identifier], tables

    def repr_(self):
        return self.identifier

    def add_fold(self, fold: str) -> Select:
        return self.filter(self.fold == fold)

    def outputs(self, **kwargs):
        """
        This method returns a query which joins a query with the model outputs.

        :param model: The model identifier for which to get the outputs

        >>> q = t.filter(t.age > 25).outputs('model_name', db)

        The above query will return the outputs of the `model_name` model
        with t.filter() ids.
        """
        return IbisCompoundSelect(
            table_or_collection=self, query_linker=self._get_query_linker(members=[])
        ).outputs(**kwargs)

    @property
    def id_field(self):
        return self.primary_id

    @property
    def select_table(self) -> Select:
        return self

    @property
    def select_ids(self) -> Select:
        return self.select(self.primary_id)

    def select_using_ids(self, ids: t.Sequence[t.Any]) -> Select:
        return self.filter(self[self.primary_id].isin(ids))

    def select_ids_of_missing_outputs(
        self, key: str, model: str, version: int
    ) -> Select:
        output_table = IbisQueryTable(
            identifier=get_output_table_name(model, version),
            primary_id='output_id',
        )
        query_id = str(hash(self))
        filtered = output_table.filter(
            output_table.key == key and output_table.query_id == query_id
        )
        return self.anti_join(filtered, filtered.input_id == self[self.primary_id])

    def select_single_id(self, id):
        return self.filter(getattr(self, self.primary_id) == id)

    def __getitem__(self, item):
        return IbisCompoundSelect(
            table_or_collection=self,
            query_linker=self._get_query_linker(
                members=[IbisQueryComponent('__getitem__', type=QueryType.ATTR)]
            ),
        )(item)

    def _insert(self, documents, **kwargs):
        return IbisInsert(documents=documents, kwargs=kwargs, table_or_collection=self)

    def _get_query(
        self,
        pre_like: t.Optional[Like] = None,
        query_linker: t.Optional[QueryLinker] = None,
        post_like: t.Optional[Like] = None,
    ) -> IbisCompoundSelect:
        return IbisCompoundSelect(
            pre_like=pre_like,
            query_linker=query_linker,
            post_like=post_like,
            table_or_collection=self,
        )

    def _get_query_component(
        self,
        k,
    ):
        return IbisQueryComponent(name=k, type=QueryType.ATTR)

    def _get_query_linker(self, members) -> IbisQueryLinker:
        return IbisQueryLinker(table_or_collection=self, members=members)

    def insert(
        self,
        *args,
        **kwargs,
    ):
        return self._insert(*args, **kwargs)

    def _delete(self, *args, **kwargs):
        return super()._delete(*args, **kwargs)

    def execute(self, db):
        return db.databackend.conn.table(self.identifier).execute()

    def model_update(
        self,
        db,
        ids: t.List[t.Any],
        key: str,
        model: str,
        version: int,
        outputs: t.Sequence[t.Any],
        flatten: bool = False,
        **kwargs,
    ):
        raise NotImplementedError


def _compile_item(item, db, tables):
    if hasattr(item, 'compile') and isinstance(
        getattr(item, 'compile'), types.MethodType
    ):
        return item.compile(db, tables=tables)
    if isinstance(item, list) or isinstance(item, tuple):
        compiled = []
        for x in item:
            c, tables = _compile_item(x, db, tables=tables)
            compiled.append(c)
        return compiled, tables
    elif isinstance(item, dict):
        c = {}
        for k, v in item.items():
            c[k], tables = _compile_item(v, db, tables=tables)
        return c, tables
    return item, tables


def _get_all_tables(item):
    if isinstance(item, (IbisQueryLinker, IbisCompoundSelect)):
        return item.get_all_tables()
    elif isinstance(item, IbisQueryTable):
        return [item.identifier]
    elif isinstance(item, list) or isinstance(item, tuple):
        return sum([_get_all_tables(x) for x in item], [])
    elif isinstance(item, dict):
        return sum([_get_all_tables(x) for x in item.values()], [])
    else:
        return []


@dc.dataclass
class IbisQueryComponent(QueryComponent):
    """
    This class represents a component of an ``ibis`` query.
    For example ``filter`` in ``t.filter(t.age > 25)``.
    """

    __doc__ = __doc__ + QueryComponent.__doc__  # type: ignore[operator]

    @property
    def renamings(self):
        if self.name == 'rename':
            return self.args[0]
        elif self.name == 'relabel':
            return self.args[0]
        else:
            out = {}
            if self.args is not None:
                for a in self.args:
                    if isinstance(
                        a, (IbisCompoundSelect, IbisQueryLinker, IbisQueryComponent)
                    ):
                        out.update(a.renamings)
            if self.kwargs is not None:
                for v in self.kwargs.values():
                    if isinstance(
                        v, (IbisCompoundSelect, IbisQueryLinker, IbisQueryComponent)
                    ):
                        out.update(v.renamings)
            return out

    def repr_(self) -> str:
        """
        >>> IbisQueryComponent('__eq__(2)', type=QueryType.QUERY, args=[1, 2]).repr_()
        """
        out = super().repr_()
        match = re.match('.*__([a-z]+)__\(([a-z0-9_\.\']+)\)', out)
        symbol = match.groups()[0] if match is not None else None

        if symbol == 'getitem':
            assert match is not None
            return f'[{match.groups()[1]}]'

        lookup = {'gt': '>', 'lt': '<', 'eq': '=='}
        if match is not None and match.groups()[0] in lookup:
            out = f' {lookup[match.groups()[0]]} {match.groups()[1]}'
        return out

    def compile(
        self, parent: t.Any, db: 'Datalayer', tables: t.Optional[t.Dict] = None
    ):
        if self.type == QueryType.ATTR:
            return getattr(parent, self.name), tables
        args, tables = _compile_item(self.args, db, tables=tables)
        kwargs, tables = _compile_item(self.kwargs, db, tables=tables)
        return getattr(parent, self.name)(*args, **kwargs), tables

    def get_all_tables(self):
        out = []
        out.extend(_get_all_tables(self.args))
        out.extend(_get_all_tables(self.kwargs))
        return list(set(out))


@dc.dataclass
class IbisInsert(Insert):
    def __post_init__(self):
        if isinstance(self.documents, pandas.DataFrame):
            self.documents = [
                Document(r) for r in self.documents.to_dict(orient='records')
            ]

    def _encode_documents(self, table: Table) -> t.List[t.Dict]:
        return [r.encode(table.schema) for r in self.documents]

    def execute(self, db):
        table = db.load(
            'table',
            self.table_or_collection.identifier,
        )
        encoded_documents = self._encode_documents(table=table)
        ids = [r[table.primary_id] for r in encoded_documents]

        db.databackend.insert(
            self.table_or_collection.identifier, raw_documents=encoded_documents
        )
        return ids

    @property
    def select_table(self):
        return self.table_or_collection


class _SQLDictIterable:
    def __init__(self, iterable):
        self.iterable = iter(iterable)

    def next(self):
        element = next(self.iterable)
        return dict(element)

    def __iter__(self):
        return self

    __next__ = next


@dc.dataclass
class RawSQL(RawQuery):
    query: str
    id_field: str = 'id'

    def execute(self, db):
        cursor = db.databackend.conn.raw_sql(self.query)
        try:
            cursor = cursor.mappings().all()
            cursor = _SQLDictIterable(cursor)
            return aladinIbisResult(cursor, id_field=self.id_field)
        except Exception:
            return cursor
